/*! \file CGrepLib.c

    \brief Library of C functions to be used in conjuction with
           HuffwordLib and agrep to achieve error-lenient pattern
           matching over compressed files.

    \author Paolo Ferragina and Alessandro Tommasi, Dipartimento di
            Informatica, Pisa (Italy)

    \date March 2003

    CGrep is a C library that implements and extends what has been
    proposed in: deMoura, Navarro, Ziviani, Baeza-Yates, ACM Trans
    Info Syst, 2000. The main idea is to adopt a scan-based approach
    to search into a file compressed via a byte-aligned and tagged
    Huffword. The specialties of this compression method ensure that
    the search does not need to decompress the compressed file, but it
    may be executed directly on the compressed file via proper
    compression of the searched patterns.  Actually the search
    consists of three steps: first the complex pattern query (possibly
    involving errors, reg exp, ....) is resolved against the
    dictionary of tokens, built to compress the file. The tokens that
    satisfy the query are coded and then their codewords are searched
    directly onto the compressed file, thus avoiding its
    decompression.  This induces a twofold speedup: we pay a cost
    proportional to actually the length of the compressed file and we
    pay the cost to search for complicated patterns (like reg exp,
    errors, ....) only on the dictionary of tokens, which is small
    compared to the entire file. The additional features that we have
    implemented wrt the original paper are proximity query, arbitrary
    number of patters each with its search options, snippet
    extraction. Furthermore, we provide a library that can be adopted
    by anyone wanting to play with its features and possibly build
    more sophisticated search engines.

    For more details please have a look at the html documentation.

    This file is licensed under LGPL terms (see file LICENSE)
*/



#include <sys/types.h>  // stat, getpid
#include <sys/stat.h>   // stat
#include <sys/wait.h>   // macros to check exit status
#include <unistd.h>     // stat, getpid
#include <fcntl.h>      // open

#include <stdio.h> // fprintf

#include "CGrepLib.h"

/* used internally to represent a word hit */
typedef struct _word_hit_t {
  size_t byte_position;
  size_t word_position;
} word_hit_t;

/* Whether cw1, of length len1, is the same as cw2, of length len2  */
#define CGREP_CWMATCH(cw1, len1, cw2, len2) \
   (((len1)!=(len2))?0:(bcmp((void *) (cw1),(void *) (cw2),(len1)) == 0))

/**
 * Search for a pattern on a compressed text using agrep.
 * Agrep is invoked over the dictionary in order to find out tokens that match
 *  against the query string. The dictionary is in the format of a token per
 * line: when using options to agrep or regexp, you must remember that things
 * like: "^p.*" will match all WORDS (not lines) beginning with the letter
 * 'p'. This is because agrep is invoked over the dictionary, that actually
 * contains a token per line.
 * Once the list of matching token is found, their corresponding codewords
 * Are inserted into a hashtable, and the compressed file is scanned looking
 * for such codewords.
 *
 * @param nres pointer to an integer that will be set to the number of
 *             occurrences found. Also used as initial array size.
 * @param ctext pointer to the memory area holding the comressed file
 * @param ctext_sz size of the compressed file
 * @param pattern pattern to search for
 * @param options options to agrep (null-terminated)
 * @return an allocated array of positions within the compressed body of
 *         the file, at which codewords corresponding to matching tokens
 *         occur, NULL if nothing is found.
 */
int * CGrep_SearchPattern(int *nres, const char *ctext, size_t ctext_sz,
			  const char *pattern, char **options) {
  int text_len, compressed_body_len;
  char *compressed_body;
  Console console;
  int i = 0, j = 0;
  int ntokens;
  char *agrepOptions[CGREP_MAX_AGREPOPTIONS];
  MyHash_table tokens;
  char filter[256];
  int *positions;

  HDecompress_getBodyAndConsole((char *) ctext, ctext_sz, &text_len,
				&compressed_body, &compressed_body_len,
				&console);

#ifdef DEBUG
  printf("TEST: dc: %d, dl: %d\n", strlen(console.dictionary.content),
	 console.dictionary.length);
#endif

  i = 0;
  /* set up the options array to execvp agrep */
  agrepOptions[i++] = CGREP_AGREP_EXECUTABLE;
  // agrepOptions[i++] = "-n";

  j = 0;
  while ( options[j] && (i < (CGREP_MAX_AGREPOPTIONS - 2)) )
    agrepOptions[i++] = (char *) options[j++];

  agrepOptions[i++] = (char *) pattern;
  agrepOptions[i] = NULL;

  /* Now, use the dictionary content to perform agrep */
  MyHashtable_init(&tokens, CGREP_HT_INITSIZE);
  bzero(filter, 256);
  ntokens = CGrep_GetMatchingCW(&tokens, filter, agrepOptions, 0, &console);
  
  if ( !ntokens ) { // nothing found in the dictionary, return
    *nres = 0;
    return(NULL);
  }

  /*
     - perform GetCWOccurrences over the compressed body
  */
  positions = CGrep_GetCWOccurrences(nres, &tokens, filter, compressed_body,
				     compressed_body_len);
  MyHashtable_clear(&tokens);
  return(positions);
}

/**
 * returns an array of positions of the pattern in the compressed body
 * as a substring,
 * allowing for at most <i>errors</i> errors. The array returned is malloc'ed.
 * This function is a 'shortcut' invocation for CGrep_SearchPattern.
 *
 * @param nres pointer to an integer where the length of the array is stored,
 *             it is also used as the size of the first allocation.
 * @param pattern is the pattern to search for
 * @param ctext is the buffer holding the compressed file's data
 * @param ctext_len length of the buffer
 * @param errors number of errors allowed in the search
 * @return a malloc'ed array of positions within the body.
 */
int *
CGrep_SearchSubstring(int *nres, const char *pattern, const char *ctext,
		      size_t ctext_len, int errors)
{
  char *options[CGREP_MAX_AGREPOPTIONS];
  char erropt[3+(int) log10(errors+1)];
  int i = 0;
  
  if ( errors ) {
    sprintf(erropt, "-%d", errors);
    options[i++] = erropt;
  }
  options[i] = NULL;

  return(CGrep_SearchPattern(nres, ctext, ctext_len, pattern, options));
}

/**
 * returns the array of positions of the word in the compressed body, with at
 * most <i>errors</i> errors. The array returned is malloc'ed.
 * This function is a 'shortcut' invocation for CGrep_SearchPattern.
 * 
 * @param nres pointer to an integer where the length of the returned array
 *             is stored. Also used as initial array size.
 * @param word the word to search for
 * @param ctext the buffer holding the compressed file data (not just
 *            the body)
 * @param ctext_len buf's length
 * @param errors number of allowed errors
 * @return an array of positions, malloc'ed by the function.
 */
int *
CGrep_SearchWord(int *nres, const char *word, const char *ctext,
		 size_t ctext_len, int errors)
{
  char *options[CGREP_MAX_AGREPOPTIONS];
  char erropt[3+(int) log10(errors+1)];
  int i = 0;
  
  options[i++] = "-w"; /* ask for word match */
  if ( errors ) {
    sprintf(erropt, "-%d", errors);
    options[i++] = erropt;
  }
  options[i] = NULL;

  return(CGrep_SearchPattern(nres, ctext, ctext_len, word, options));
}

int CGrep_GetMatchingCWFromFile(const char *fname, MyHash_table *ht,
				char *filter, char **options,
				int npattern, const Console *c);

/* Invokes 'agrep' to obtain the list of separators within 'buf'. Agrep is
 *  invoked with parameters: agrep -n -v ^[0-9A-Za-z].
 *
 * @param dFileName name of the file with the dictionary
 * @param nseps pointer to an integer storing the length of the (allocated)
 *              array of the results. If != 0, it can be used to indicate
 *              the initial array size.
 * @param buf   pointer to the dictionary string
 * @param len   length of the dictionary string
 * @return an allocated array of integers indicating the positions of
 *         separators within buf
 */
int
CGrep_GetSeparators(char *dFileName, MyHash_table *ht, char *filter,
		    Console *c)
{
  char *agrepOptions[CGREP_MAX_AGREPOPTIONS];

  int i = 0;
  agrepOptions[i++] = CGREP_AGREP_EXECUTABLE;
  // agrepOptions[i++] = "-n";
  agrepOptions[i++] = "-v";
  // agrepOptions[i++] = "-i";
  agrepOptions[i++] = "^[0-9A-Za-z]";
  agrepOptions[i] = NULL;

  /* Now, use the dictionary content to perform agrep */
  return CGrep_GetMatchingCWFromFile(dFileName, ht, filter, agrepOptions, -1,
				     c);
}

void CGrep_WriteDictToFile(char *fname, int len, const Dictionary *d);


/** Performs a proximity search, returning an array of proximity_hit_t's
 *  Up to CGREP_MAX_PATTERNS (10) patterns are allowed, with all of agrep's
 *  options (regexp search, approximated search, case
 *  sentitiveness/insensitiveness, etc.) In case of non exact search, it
 *  may happen that a word matches multiple words. In this case,
 *  if a set of words matches the search in multiple ways, it is returned
 *  only once (the arrays <i>positions</i> and <i>ranks</i> in the
 *  corresponding proximity_hit_t will hold values for the first such
 *  match found.
 *
 * @param nres pointer to an integer that will hold the number of elements
 *        in the (allocated) array of results. If != 0 on call, it is used
 *        as the initial array size.
 * @param ctext buffer holding the mmapped compressed file
 * @param ctext_sz size of the compressed file
 * @param prox_window width of the proximity window
 * @param patterns array of strings holding the patterns to search for. The
 *        array must be NULL-terminated.
 * @param options array of array of options to agrep (one array of options
 *        per pattern). Must be NULL-terminated in both dimensions.
 * @return an allocated array of proximity_hit_t, NULL if no results are
 *         found.
 */ 
proximity_hit_t *
CGrep_SearchProximity(int *nres, const char *ctext, size_t ctext_sz,
		      int prox_window, char **patterns,
		      char ***options)
{
  int text_len, compressed_body_len;
  char *compressed_body;
  Console console;
  int correction = 0, i = 0, j = 0;
  int ntokens;
  int npatterns;
  int nseparators;
  char *agrepOptions[CGREP_MAX_AGREPOPTIONS];
  MyHash_table tokens;
  MyHash_table sepTable;
  Hash_node *newline_node;
  char filter[256];
  char sepFilter[256];
  char dFileName[100];
  proximity_hit_t *ret;

  HDecompress_getBodyAndConsole((char *) ctext, ctext_sz, &text_len,
				&compressed_body, &compressed_body_len,
				&console);
  
  MyHashtable_init(&tokens, CGREP_HT_INITSIZE);
  bzero(filter, 256);
  bzero(sepFilter, 256);

#ifdef DEBUG
  printf("TEST: dc: %d, dl: %d\n", strlen(console.dictionary.content),
	 console.dictionary.length);
#endif

  /* Is one of the dictionary tokens "\n"??? */
  if ( (newline_node = HHashtable_search("\n", 1, &(console.hashtable))) )
    HToken_RankFromTaggedCw(newline_node->codeword, newline_node->cw_len,
			    &(console.canoinfo),
			    &correction);
  else
    correction = -1;

  CGrep_WriteDictToFile(dFileName, 100, &(console.dictionary));

  /* for each pattern, invoke agrep and fill the relative hashmap */
  for ( npatterns = 0; patterns[npatterns]; npatterns++ ) {
    i = 0;
    /* set up the options array to execvp agrep */
    agrepOptions[i++] = CGREP_AGREP_EXECUTABLE;
    // agrepOptions[i++] = "-n";
    
    j = 0;
    while ( options[npatterns] && options[npatterns][j] &&
	    (i < CGREP_MAX_AGREPOPTIONS) )
      agrepOptions[i++] = options[npatterns][j++];

    agrepOptions[i++] = patterns[npatterns];
    agrepOptions[i] = NULL;

    /* Now, use the dictionary content to perform agrep */
    ntokens = CGrep_GetMatchingCWFromFile(dFileName, &tokens, filter,
					  agrepOptions, npatterns, &console);
    if ( !ntokens ) { // if no tokens are found for this pattern, we can return
      *nres = 0;
      return(NULL);
    }
  }

  /* find all separators */
  MyHashtable_init(&sepTable, 10*CGREP_HT_INITSIZE); /* TEN times larger?! */
  nseparators = CGrep_GetSeparators(dFileName, &sepTable, sepFilter, &console);

  /* done with the dictionary file, unlink it */
  unlink(dFileName);

  /*
    I now have the list of codewords corresponding to the patterns
    and to the separators. Now let's go find the occurrences...
  */

  if ( *ctext ) // spaceless model
    ret = CGrep_GetOccurrencesProximitySpaceless(nres, prox_window, npatterns,
						 &tokens, filter, &sepTable,
						 sepFilter,
						 compressed_body,
						 compressed_body_len);
  else
    /* the file keeps spaces as separators; optimize */
    ret = CGrep_GetOccurrencesProximity(nres, prox_window,
					npatterns, &tokens,
					filter, &sepTable,
					sepFilter,
					compressed_body,
					compressed_body_len,
					newline_node);

  MyHashtable_clear(&tokens);
  MyHashtable_clear(&sepTable);
  
  return(ret);
}

/* Dumps the dictionary to a file, whose name is stored in fname.
   len is the length of fname's buffer */
void
CGrep_WriteDictToFile(char *fname, int len, const Dictionary *d)
{
  char pids[100]; /* actually pids have at most 5 chars... */
  int attempt = 0;
  int tmpfd;
  int written = 0;

  if ( len < 40 ) {
    fputs("Too small a buffer top hold the filename.", stderr);
    exit(2);
  }

  sprintf(pids, "%d", getpid());

  do
    sprintf(fname, "/tmp/agrep.tmp.%s.%d", pids, ++attempt);
  while ( ((tmpfd = open(fname, O_RDWR|O_CREAT|O_EXCL,
			 S_IREAD|S_IWUSR)) == -1) &&
	  (errno == EEXIST) );
    
  if ( tmpfd == -1 ) {
    perror("open");
    exit(2);
  }
  
  do {
    written =  write(tmpfd, d->content + written, d->length - written);
    if ( written == -1 ) {
      perror("write");
      exit(2);
    }
  } while ( written < d->length );
  close(tmpfd);
}

int
CGrep_Lookup(const char*ifname, const char*ofname, char **options)
{
  char cmd[10000];
  int i;
  int status;

  strcpy(cmd, options[0]);
  strcat(cmd, " ");
  for (i = 1; options[i]; i++ ) {
    strcat(cmd, "'");
    strcat(cmd, options[i]);
    strcat(cmd, "' ");
  }
  
  strcat(cmd, ifname);
  strcat(cmd, " > ");  
  strcat(cmd, ofname);

#ifdef DEBUG
  puts(cmd);
#endif
  status = system(cmd);
  if ( (status == -1) ) { // system failed
    perror("system");
    exit(2);
  }
  if ( !WIFEXITED(status) ) {
    fprintf(stderr, "command \"%s\" failed.\n", cmd);
    exit(2);
  }
  switch ( WEXITSTATUS(status) ) {
  case 0: // everything normal (matches found)
    break;
  case 1: // ok, there were simply no matches
    break;
  default: // exit status not 0 and not 1: something failed
    fprintf(stderr, "system(\"%s\") exited with status %d\n",
	    cmd, WEXITSTATUS(status));
    exit(2);
  }

  return(0);
}

/* Actually implements GetMatchingCW */
int CGrep_GetMatchingCWFromFile(const char *fname, MyHash_table *ht,
				char *filter, char **options,
				int npattern, const Console *c)
{  
  int i = 0;
  char pids[100] = "";
  FILE *istream;
  char obufs[1024];
  int attempt = 0;
  char *input_buf, *runner, *buf_end;
  size_t input_len;
  struct stat sbuf;
  int ntok = 0;
  int newLineSeen = 0;

  sprintf(pids, "%d", getpid());
  sprintf(obufs, "/tmp/agrep.out.%s.%d", pids, attempt);

  CGrep_Lookup(fname, obufs, options);

  stat(obufs, &sbuf);
  if ( (istream = fopen(obufs, "r")) == NULL ) {
    perror("fopen");
    exit(2);
  }
  input_len = sbuf.st_size;

  if ( input_len ) {
    input_buf = (char *) mmap(NULL,input_len,PROT_READ,MAP_SHARED,
			      fileno(istream),0);
    if ( (input_buf == MAP_FAILED) ) {
      perror("mmap");
      exit(2);
    }
    if ( input_buf == NULL ) {
      fprintf(stderr, "mmap returned NULL (but the filesize was >0)?!\n");
      exit(2);
    }
    
    
    buf_end = input_buf + input_len - 1;
    runner = input_buf;
    ntok = 0;
    do {
      char *end = runner;
      Hash_node *t;
      int hashcw;
      while ( *end != '\n' ) end++;
      if ( (end == runner) && !newLineSeen ) {
	while ( *(++end) == '\n' );
	end -= 1;
	if ( end - runner != 1 ) {
	  fprintf(stderr,
		  "Error: messed up dictionary? (incoherent '\\n', line %d)\n",
		  ntok);
	  exit(2);
	}
	// newLineSeen = 1;
      } else if ( end == runner ) {
	fprintf(stderr,
		"Error: messed up dictionary? (incoherent '\\n', line %d)\n",
		ntok);
	exit(2);
      }
      ntok++;
      t = HHashtable_search(runner, (end - runner),
			    (HHash_table *) &(c->hashtable));
      if ( !t ) { // something must have gone REALLY wrong, here...
	char wrongToken[end-runner+1];
	memcpy(wrongToken, runner, (end-runner));
	wrongToken[end-runner] = '\0';
	fprintf(stderr,
		"Error: messed up dictionary? (token \"%s\" not found)\n",
		wrongToken);
	exit(2);
      }
#ifndef WORDS_BIGENDIAN
      /* revert the bytes in the codeword (if little-endian) */
      for ( i = 0; i < t->cw_len; i++ )
	*((char *) &hashcw + i) = *((char *) &t->codeword + t->cw_len - i - 1);
#endif

      if ( filter )
	filter[*((unsigned char *) &hashcw)] = 1;
      
      MyHashtable_insert((char *) &hashcw, t->cw_len, npattern, ht
#ifdef DEBUG
			 , runner, (end - runner)
#endif
			 );

      runner = end + 1;
    } while ( runner < buf_end );
    munmap(input_buf, input_len);
  } // if ( input_len )
  fclose(istream);
#ifndef DEBUG
  unlink(obufs);
#else
  printf("N. results: %d\n", ntok);
#endif

  return(ntok);
}

/** 
 * Fills the hashtable <i>ht</i> with the codewords corresponding to all
 * the words in the Dictionary obtained from Console <i>c</i> matching the
 * search pattern as specified in agrep's <i>options</i>.
 * 'Systems' agrep, I/O via file. The dictionary content is saved to
 * '/tmp/agrep.tmp.PID.n', and the output of agrep to '/tmp/agrep.out.PID.n'.
 * Temporary files are unlink'ed after the execution. "-n" or other
 * flags that modify agrep's output format MUST NOT be within
 * the options, and the buffer must contain a word per line. <i>options[0]</i>
 * must be agrep's executable name, and the array must be null-terminated.
 * Along with the hashtable (which must be initialized), if <i>filter</i> is
 * not NULL, it is interpreted as an array of 256 unsigned char to fill with
 * 1's in correspondance to the fisrt byte of every matching codeword found.
 *
 * @param ht pointer to a MyHash_table that is filled with the codewords
 *           corresponding to the tokens matching the search pattern
 * @param filter if not NULL, this array of 256 unsigned chars is filled
 *               with 1's in correspondance with the first byte of every
 *               codeword inserted in <i>ht</i>
 * @param options options passed to agrep (including the search pattern and
 *                the executable name)
 * @param npattern id of the search pattern. Used to distinguish among
 *                 matches against different patterns in proximity search
 *                (each pattern should have a different id)
 * @param c pointer to the console of the compressed file
 * @return the number of codewords inserted in <i>ht</i>.
 */
int
CGrep_GetMatchingCW(MyHash_table *ht, char *filter, char **options,
		    int npattern, const Console *c)
{
  char fname[100];
  int ret;

  CGrep_WriteDictToFile(fname, 100, &(c->dictionary));
  ret = CGrep_GetMatchingCWFromFile(fname, ht, filter, options, npattern, c);
  unlink(fname);
  return(ret);
}

/**
 * Looks for the codewords contained within the hashtable in the compressed
 * body. Returns the list of positions (in the compressed body)
 * of the matches.
 *
 * @param nocc pointer to an integer that is filled with the length of
 *             the hitlist
 * @param ht pointer to a MyHash_table holding the (tagged) codewords
 * @param filter pointer to an array of 256 elements holding 1 corresponding
 *        to the positions of the first byte of codewords in ht.
 * @param cbody buffer holding the compressed body
 * @param cbody_len length of the compressed body
 * @return an allocated array of positions within the compressed body.
 **/
int *
CGrep_GetCWOccurrences(int *nocc, const MyHash_table *ht, const char *filter,
		       const char *cbody, size_t cbody_len)
{
  int *position;
  const char *prevpos, *currpos = cbody;
  int currentResSize = 0;
  const char *endpos = cbody+cbody_len;
  register char v;

  currentResSize = CGREP_MAX(*nocc,CGREP_INIT_VECSIZE);
  position = (int *) malloc(currentResSize * sizeof(int));

  *nocc = 0;
  while ( currpos < endpos ) {
    prevpos = currpos;
    v = *(*((unsigned char *) prevpos) + filter);

    do {
      currpos++;
    } while ( (currpos<endpos) && ((*currpos & 0x80) == 0) );
    
    if ( v ) {
      if ( MyHashtable_search(prevpos, currpos-prevpos, ht) ) {
	/* we have found a new hit, record its position */
	
	if ( *nocc >= currentResSize ) {
	  /* must reallocate the res array */
	  currentResSize += (int) sqrt(currentResSize) + 1;
	  position = (int *) realloc(position, currentResSize * sizeof(int));
	}
	
	position[(*nocc)++] = (int) (prevpos-cbody);
      }
    }
  }
  return(position);
}

/**
 * Looks for the codewords contained in the hashtable <i>ht</i> in the
 * compressed body <i>cbody</i>. Returns the position of the first match
 * found.
 *
 * @param len pointer to an integer that is filled with the length of
 *             the matching codeword
 * @param ht pointer to a Hashtable holding the (tagged) codewords
 * @param filter pointer to an array of 256 elements holding 1 corresponding
 *        to the positions of the first byte of codewords in ht.
 * @param cbody buffer holding the compressed body
 * @param remaining characters remaining to the end of the buffer
 * @return a pointer to the first occurrence of a codeword, NULL if there are
 *         no occurrences
 **/
const char *
CGrep_GetNextCWOccurrence(int *len, MyHash_table *ht, const char *filter,
			  const char *cbody, size_t remaining)
{
  const char *endpos = cbody+remaining;
  const char *prevpos;
  register char v;

  while ( cbody < endpos ) {
    prevpos = cbody;
    v = *(*((unsigned char *) prevpos) + filter);

    do {
      cbody++;
    } while ( (cbody<endpos) && ((*cbody & 0x80) == 0) );
    
    if ( v ) {
      if ( MyHashtable_search(prevpos, cbody-prevpos, ht) ) {
	/* we have found a new hit, return it */
	*len = cbody-prevpos;
	return(prevpos);
      }
    }
  }
  return(NULL);
}

/* Macro to move to the beginning of the next codeword. WARNING: uses
   open variables! */
#define CGREP_GET_NEXT_CW() { \
    prevpos = currpos; \
    do { \
      currpos++; \
    } while ( (currpos<endpos) && ((*currpos & 0x80) == 0) ); \
    cw_len = currpos - prevpos; \
}


/* Checks whether cw, of length cw_len is a separator, checking first
   in the filter, and then on the hashtable */
#define CGREP_CHECKIFSEPARATOR(cw, cw_len, filter, ht) \
   ((filter[*((unsigned char *) cw)]) && MyHashtable_search(cw, cw_len, ht))

/**
 * returns an array of proximity_hit_t, proximity hits for the patterns
 * over the compressed body of a spaceless-compressed text.
 * The array returned is malloc'ed; each proximity_hit_t entry is composed
 * of:
 *   * byte_position, the position in the compressed body of the beginning
 *                    of the matching window
 *   * start_position, the rank of the first word of the matching window
 *   * end_position, the rank of the last word in the matching window
 *   * positions, an array of positions in the cbody of the match
 *                for each pattern
 *   * ranks, an array of ranks giving the rank of each pattern
 * 
 * @param nocc pointer to an integer there the length of the array is stored
 * @param prox_window width of the proximity window
 * @param npatterns number of patterns sought after;
 * @param ht pointer to a hashtable holding the codewords corresponding to
 *           the patterns
 * @param filter this array must contain 1's in correspondance to the first
 *               byte of each codeword in <i>ht</i>. Must be full of 1's
 *               if a filter is not to be applied.
 * @param separators pointer to a hashtable holding the codewords
 *                   corresponding to the separators
 * @param sepFilter this array must contain 1's in correspondance with the
 *                  first byte of each codeword in <i>separators</i>. Must
 *                  be full of 1's if a filter is not to be applied.
 * @param cbody the compressed body to search onto
 * @param cbody_len the length of the compressed body
 * @return an allocated array of proximity_hit_t, NULL if no results are found.
 */
proximity_hit_t *
CGrep_GetOccurrencesProximitySpaceless(int *nocc, int prox_window,
				       int npatterns,
				       const MyHash_table *ht,
				       const char filter[],
				       const MyHash_table *separators,
				       const char sepFilter[],
				       const char *cbody, size_t cbody_len)
{
  proximity_hit_t *matches;
  word_hit_t patternHit[npatterns];    
  int i, k;
  int patternIndexWithMinPosition;
  size_t minWPosition, minBytePosition;
  size_t currWPosition = 0;
  const char *prevpos, *currpos = cbody;
  int currentResSize = 0;
  const char *endpos = cbody+cbody_len;
#ifdef DEBUG
  int nseps = 0;
#endif
  int cw_len = 0;
  MyHash_node *hn;

  currentResSize = CGREP_MAX(*nocc,CGREP_INIT_VECSIZE);
  matches = (proximity_hit_t *) malloc(currentResSize *
				       sizeof(proximity_hit_t));

  for (i = 0; i < npatterns; i++ ) {
    patternHit[i].byte_position = 0;
    patternHit[i].word_position = 0;
  }

  *nocc = 0;
  patternIndexWithMinPosition = 0;
  minWPosition = 0;
  minBytePosition = 0;

  while ( currpos < endpos ) {
    CGREP_GET_NEXT_CW();
    
    /* first, check if it's a separator */
    if ( sepFilter[*(unsigned char *) prevpos] ) {
      if ( MyHashtable_search(prevpos, cw_len, separators) ) {
	/* it's a separator, skip! */
#ifdef DEBUG
	nseps++;
#endif
	continue;
      }
    }
    
    currWPosition++;

    hn = NULL;
    while ( (hn = CGrep_CheckIfIsPattern(prevpos, cw_len,
					 &(ht[0]), hn, filter)) != NULL ) {
      i = hn->npattern;
      /* found a hit for the i-th pattern */
#ifdef DEBUG
      printf("Found a hit for pattern %d, position %d\n", i, currWPosition);
#endif
      patternHit[i].word_position = currWPosition;
      patternHit[i].byte_position = (int) (prevpos-cbody);
      if ( patternIndexWithMinPosition == i ) {
	/* find the new min */
	int j = 0;
	
	minWPosition = currWPosition;	    
	for ( j = 0; j < npatterns; j++ ) {
	  if ( minWPosition >= patternHit[j].word_position ) {
	    patternIndexWithMinPosition = j;
	    minWPosition = patternHit[j].word_position;
	    minBytePosition = patternHit[j].byte_position;
	  }
	}
      }
      
      /* see if we are within the proximity window (but only if
	 all patterns have a match */
      if ( (minWPosition > 0) &&
	   ((currWPosition - minWPosition) <= (size_t) prox_window) ) {
	/* found a new proximity hit! */
#ifdef DEBUG
	printf("...and a proximity hit!\n");
#endif
	if ( *nocc >= currentResSize ) {
	  /* must reallocate the res array */
	  currentResSize += (int) sqrt(currentResSize) + 1;
	  matches = (proximity_hit_t *) realloc(matches,
						currentResSize *
						sizeof(proximity_hit_t));
	}
	
	matches[*nocc].byte_position = minBytePosition;
	matches[*nocc].start_position = minWPosition;
	matches[*nocc].end_position = currWPosition;
	for ( k = 0; k < npatterns; k++ ) {
	  matches[*nocc].positions[k] = patternHit[k].byte_position;
	  matches[*nocc].ranks[k] = patternHit[k].word_position;
	}
	(*nocc)++;
      }	  
    }
  }
  
  /* if there were no matches, free the memory and return NULL */
  if ( !(*nocc) ) {
    free(matches);
    matches = NULL;
  }

  return(matches);
}

/*  Used to encode states in the automa used in
 *  CGrep_GetOccurrencesProximity.
 */
typedef enum state {FOUND_WORD, FOUND_NEWLINE, FOUND_SEPARATOR} state_t;

/**
 *  This function is the same as CGrep_GetOccurrencesProximitySpaceless, but
 *  is intended for files compressed with spaces.
 *  Has an additional argument: a pointer to the hashnode
 *  relative to the newline entry in the dictionary (NULL if there is no
 *  newline).
 */
proximity_hit_t *
CGrep_GetOccurrencesProximity(int *nocc, int prox_window,
			      int npatterns, const MyHash_table *ht,
			      const char filter[],
			      const MyHash_table *separators,
			      const char *sepFilter,
			      const char *cbody, size_t cbody_len,
			      const Hash_node *nl)
{
  proximity_hit_t *matches;
  word_hit_t patternHit[npatterns];    
  int i;
  int patternIndexWithMinPosition;
  size_t minWPosition, minBytePosition;
  size_t currWPosition = 0;
  const char *prevpos, *currpos = cbody;
  int currentResSize = 0;
  const char *endpos = cbody+cbody_len;
  int cw_len = 0;
  int newline_cw = 0;
  int newline_cw_len;
  state_t state;
  MyHash_node *hn;
  int k;

  if ( !nl )
    newline_cw_len = 0; /* this way all matches against '\n' will return 0 */
  else {
    newline_cw = nl->codeword;
    newline_cw_len = nl->cw_len;
  }

  currentResSize = CGREP_MAX(*nocc,CGREP_INIT_VECSIZE);
  matches = (proximity_hit_t *) malloc(currentResSize *
				       sizeof(proximity_hit_t));

  for (i = 0; i < npatterns; i++ ) {
    patternHit[i].byte_position = 0;
    patternHit[i].word_position = 0;
  }

  *nocc = 0;
  patternIndexWithMinPosition = 0;
  minWPosition = 0;
  minBytePosition = 0;

  CGREP_GET_NEXT_CW();

  state = FOUND_NEWLINE; /* safe state to begin with */

  do {
    switch ( state ) {

    case FOUND_SEPARATOR:
      /* first, check if we have a newline */
      if ( CGREP_CWMATCH(prevpos, cw_len, &newline_cw, newline_cw_len) ) {
	state = FOUND_NEWLINE;
	CGREP_GET_NEXT_CW();
	break;
      }
      state=FOUND_WORD;
      // break; // no need to break, can go straight to FOUND_WORD
      
    case FOUND_WORD:
      currWPosition++;

      hn = NULL;
      
      while ( (hn = CGrep_CheckIfIsPattern(prevpos, cw_len,
					   ht, hn, filter)) != NULL ) {
	i = hn->npattern;
	patternHit[i].word_position = currWPosition;
	patternHit[i].byte_position = (int) (prevpos-cbody);
	if ( patternIndexWithMinPosition == i ) {
	  /* find the new min */
	  int j = 0;
	  
	  minWPosition = currWPosition;	    
	  for ( j = 0; j < npatterns; j++ ) {
	    if ( minWPosition >= patternHit[j].word_position ) {
	      patternIndexWithMinPosition = j;
	      minWPosition = patternHit[j].word_position;
	      minBytePosition = patternHit[j].byte_position;
	    }
	  }
	}
	
	/* see if we are within the proximity window (but only if
	   all patterns have a match */
	if ( (minWPosition > 0) &&
	     ((currWPosition - minWPosition) <= (size_t) prox_window) ) {
	  /* found a new proximity hit! */
	  if ( *nocc >= currentResSize ) {
	    /* must reallocate the res array */
	    currentResSize += (int) sqrt(currentResSize) + 1;
	    matches = (proximity_hit_t *) realloc(matches,
						  currentResSize *
						  sizeof(proximity_hit_t));
	  }
	  
	  matches[*nocc].byte_position = minBytePosition;
	  matches[*nocc].start_position = minWPosition;
	  matches[*nocc].end_position = currWPosition;
	  for ( k = 0; k < npatterns; k++ ) {
	    matches[*nocc].positions[k] = patternHit[k].byte_position;
	    matches[*nocc].ranks[k] = patternHit[k].word_position;
	  }
	  (*nocc)++;
	}
      }
      
      CGREP_GET_NEXT_CW();
      if ( CGREP_CWMATCH(prevpos, cw_len, &newline_cw, newline_cw_len) ) {
	state = FOUND_NEWLINE;
	CGREP_GET_NEXT_CW();
	break;
      }
      state = FOUND_SEPARATOR;
      CGREP_GET_NEXT_CW();
      break;
      
    case FOUND_NEWLINE:
      /* first, check if we have another newline */
      if ( CGREP_CWMATCH(prevpos, cw_len, &newline_cw, newline_cw_len) ) {
	state = FOUND_NEWLINE;
	CGREP_GET_NEXT_CW();
	break;
      }
      /* otherwise, check to see if we have a separator */
      if ( CGREP_CHECKIFSEPARATOR(prevpos, cw_len, sepFilter, separators) ) {
	state=FOUND_SEPARATOR;
	CGREP_GET_NEXT_CW();
	break;
      }
      /* otherwise, we have a word for sure: go to the relevant state */
      state=FOUND_WORD;
      break;
    }
  } while ( (currpos < endpos) );
  
  if ( state == FOUND_WORD ) {
    /* Check the last word, but warning: it might have been checked already */
  }

  /* Question: is the LAST codeword found alright?!?! */

  if ( !(*nocc) ) { // free matches and return NULL
    free(matches);
    matches = NULL;
  }

  return(matches);
}

/** Checks whether a codeword matches with one of the patterns sought for.
 *  <i>ht</i> must hold the codewords relative to the patterns. Because 
 *  the same codeword can represent a word that matches more than one pattern,
 *  the function allows to be called repeatedly. If the parameter <i>hn</i>
 *  is NULL, the first match is returned. If <i>hn</i> holds the value
 *  previously returned, the next one is returned.
 *
 * @param cbody pointer to the beginning of the codeword in the compressed body
 * @param cw_len length of the codeword
 * @param ht hashtable holding the codewords sought for
 * @param hn pointer to the previously returned value, must be NULL for
 *           the first call
 * @param filter array of 256 elements holding 1's in correspondance with the
 *               first byte of codewords in ht (use a array of all 1's
 *               if you don't want to filter over the first byte)
 * @return pointer to the next MyHash_node holding the match, NULL if no
 *         match is found
 */
MyHash_node *
CGrep_CheckIfIsPattern(const char *cbody, int cw_len, const MyHash_table *ht,
		       MyHash_node *hn, const char filter[])
{
  if ( !hn ) {
    if ( filter[*(unsigned char *) cbody] ) {
      // first, use the filters
      hn = ht->table[MyHashtable_func(cbody,cw_len,ht)];
      while ( hn && 
	      !CGREP_CWMATCH(cbody, cw_len, hn->str, hn->len_str) )
	hn = hn->next;
    } else
      return(NULL);
  } else { // keep scanning the hashtable
    do {
      hn = hn->next;
    } while ( hn && 
	      !CGREP_CWMATCH(cbody, cw_len, hn->str, hn->len_str) );
  }
  return(hn);
}

/** Utility function: given a string, it copies it to another allocated string
 *  where some of the characters are escaped as: [\<hex value>], where
 *  <hex value> is the hexadecimal value of the escaped character.
 *  The characters to escape are selectable by range and list: all
 *  character whose value is smaller than <i>min</i>, greater than
 *  <i>max</i>, or equal to one in <i>exceptions</i> are escaped.
 *
 * @param s the string to be escaped
 * @param len the length of s
 * @param min minimum printable character
 * @param max maximum printable character
 * @param exceptions null-terminated array of characters to escape
 * @return an allocated, escaped version of s
 */
char *CGrep_escapeStringConfigurable(const char *s, size_t len, char min,
				     char max, const char *exceptions)
{
  int sz = 5 * len + 1;
  char *ret = (char *) malloc(sz);
  const char *end = s+len;
  const char *bgn = ret;
  char *curr = ret;
  if ( !ret ) {
    perror("malloc");
    exit(2);
  }
  while ( s < end ) {
    if ( (curr - bgn) >= (sz-5) ) { // resize
      sz = (sz * 0.5) + 5;
      ret = realloc(ret, sz);
      if ( !ret ) {
	perror("realloc");
	exit(2);
      }
      curr = ret + (curr - bgn);
      bgn = ret;
    }
    if ( (*s < min) || (*s > max) || strchr(exceptions, *s) ) {
      sprintf(curr, "[\\%x]", *s);
      while ( *curr ) curr++;
    } else
      *curr++ = *s;
    s++;
  }
  *curr = 0;
  return(ret);
}

/** "Default" version of CGrep_escapeStringConfigurable, using the defaults
 *  CGREP_MIN_PRINTABLE_CHAR (32), CGREP_MAX_PRINTABLE_CHAR (126), and
 *  CGREP_NONPRINTABLE_CHARS ("[]"). 
 */
char *CGrep_escapeString(const char *s, size_t len)
{
  return(CGrep_escapeStringConfigurable(s, len, CGREP_MIN_PRINTABLE_CHAR,
					CGREP_MAX_PRINTABLE_CHAR,
					CGREP_NONPRINTABLE_CHARS));
}


/* ---------------------------------------------------- */
/* -------------------  MyHash Table    ----------------- */
/* ---------------------------------------------------- */

/** Initialize the hash table according to the number of estimated
 *   tokens; the load factor is set to 0.1.
 *
 * @param ht pointer to an (empty) hash table.
 * @param n estimated number of items to be inserted.
 */
void MyHashtable_init(MyHash_table *ht, int n)
{
  int i;
  ht->size  = n * 10;     // Load factor 0.1
  ht->table = (MyHash_nodeptr_array) malloc(ht->size * sizeof(MyHash_node*));
  ht->card = 0;
  if (ht->table == NULL) {
    fprintf(stderr,"Fatal Error: MyHash table allocation\n");
    exit(2);
  }

  /* memset(ht->table, 0, ht->size * sizeof(MyHash_node)); */
  /* A cleaner way (albeit slower): */
  for ( i = 0; i < ht->size; ht->table[i++] = NULL );
}

/** Frees all elements of a hashtable. After this call, ht is an empty,
 * uninitialized MyHash_table.
 * 
 * @param ht pointer to a MyHash_table
 */
void MyHashtable_clear(MyHash_table *ht)
{
  int i;

  for ( i = 0; i < ht->size; i++ ) {
    MyHash_node *hn = ht->table[i];
    while ( hn ) {
      MyHash_node *toFree = hn;
      hn = hn->next;
      free(toFree);
    }
  }
  ht->card = 0;
  ht->size = 0;
  free(ht->table);
}

/** Computes the hash value for the given string.
 * @param s input string whose hash value has to be computed.
 * @param len length of s.
 * @param ht pointer to an hash table. 
 */
int MyHashtable_func(const char *s, int len, const MyHash_table *ht)
{
  register int hfn;
  int hfi;
  int table_size = ht->size;

  hfn = 11;
  for (hfi=0; hfi<len ; hfi++)
    hfn = hfn ^ ((hfn<<5) + (hfn>>2) + (unsigned char) *s++);
  hfn = abs(hfn % table_size);
  return(hfn);
}

/** Searches for the given string into the passed hash table (NULL if not).
 *  
 * @param s string to be searched.
 * @param slen length of s (to manage also the NULL char).
 * @param ht pointer to the hash table to be searched.
 */
MyHash_node *MyHashtable_search(const char *s, int slen,
				const MyHash_table *ht)
{   
  MyHash_node *ghsp;

  ghsp = ht->table[MyHashtable_func(s,slen,ht)];

  while ( ghsp ) {
    if ( CGREP_CWMATCH(s, slen, ghsp->str, ghsp->len_str) )
      return(ghsp);
    ghsp = ghsp->next;
  }
  return(NULL);
}

/** Inserts the token in the hash table and returns 1 if new, 0
 *  otherwise; it also updates the counter of occurrences for that
 *  token. 
 *
 * @param s string to be inserted.
 * @param slen length of s (to manage also the NULL char, WARNING slen MUST
 *             be smaller than 4).
 * @param npattern pattern ID, useful for proximity search
 * @param ht pointer to the hash table.
 */
int MyHashtable_insert(const char *s, int slen, int npattern, MyHash_table *ht
#ifdef DEBUG
		       , char *word, int len
#endif
		       )   
     
{   
  int hiv;
  int i;
  MyHash_node *hip;
  MyHash_node *table_head;
  MyHash_node *hip_pred;


  hiv = MyHashtable_func(s,slen,ht);  // compute the hash value
  hip = MyHashtable_search(s,slen,ht);   // check string occurrence

  if (hip) {
    
    // move-to-front the searched token
    if (ht->table[hiv] != hip){

      // Find hip predecessor which surely does exist
      // because hip is not the head of the list
      for (hip_pred = ht->table[hiv]; hip_pred->next != hip; 
	   hip_pred = hip_pred->next) ;

      hip_pred->next = hip->next;   // jump hip in the old list
      table_head = ht->table[hiv];   // save the old head of the list-entry
      ht->table[hiv] = hip;        // hip is the new head of the list
      hip->next = table_head;         
    }

    return(0);

  } else {    //------ The token is new -------
    hip = (MyHash_node *) malloc(sizeof(MyHash_node));
    if (hip == NULL) {
      fprintf(stderr,"Error: Insert hash table\n");
      exit(2);
    } 

    hip->len_str = slen;
    for ( i = 0; i < slen; i++ )
      hip->str[i] = s[i];
    hip->npattern = npattern;
    hip->next = ht->table[hiv];
#ifdef DEBUG
    strncpy(hip->word, word, len);
    hip->word[len] = 0;
#endif
    ht->table[hiv] = hip;
    ht->card += 1;
    return(1);
  }
}

