/** \file CGrepLib.h
 * \brief header file with prototypes of the functions defined in
 *   CGrepLib.c
 *
 * \author Paolo Ferragina and Alessandro Tommasi, Dipartimento di
 *           Informatica, Pisa (Italy)
 *    
 * \date March 2003
 *
 * Released under LGPL (see LICENSE file for license terms)
 */ 

#include <HuffwordLib.h>

#define CGREP_MAX(a, b) (((a)>(b))?(a):(b))

#define CGREP_MAX_AGREPOPTIONS (100)
#define CGREP_INIT_VECSIZE  (100)

#ifndef CGREP_AGREP_EXECUTABLE
#define CGREP_AGREP_EXECUTABLE "agrep"
#endif

#ifndef CREP_HT_INITSIZE
#define CGREP_HT_INITSIZE (1000)
#endif

#define CGREP_MAX_PATTERNS 10

#define CGREP_MIN_PRINTABLE_CHAR 32
#define CGREP_MAX_PRINTABLE_CHAR 126
#define CGREP_NONPRINTABLE_CHARS "[]"

/** struct representing a proximity hit
 */
typedef struct _proximity_hit_t {
  size_t byte_position;  //!< position of the beginning of the prox. hit
  size_t start_position; //!< word rank of the beginning of the hit
  size_t end_position;   //!< word rank of the end of the hit
  int positions[CGREP_MAX_PATTERNS];   //!< array of positions for each pattern
  int ranks[CGREP_MAX_PATTERNS];       //!< array of ranks for each patterns
} proximity_hit_t;

/*! Elements of the hash table.
    The key (at most 4 bytes, this hashtable is used to contain codewords)
    is stored in <b>str</b>, its length in <b>len_str</b>. <b>npattern</b>
    is an id used to differentiate entries when they are relative to different
    proximity words.
 */ 
typedef struct MyHash_node {
  char str[4];         //!< token
  int len_str;         //!< length of the token (to manage also NULL)
  int npattern;        //!< id of the pattern (for the proximity search)
  struct MyHash_node *next; //!< pointer to the next element
#ifdef DEBUG
  char word[100];
#endif
  //  struct MyHash_node *next;    //!< pointer to the next item in the list
} MyHash_node;

typedef MyHash_node **MyHash_nodeptr_array; //!< array of Hash_nodes lists

/*! Data structure for an hash table managed by catenating lists and
  MTF. The field "size" indicates the table size, whereas the field
  "card" indicates the number of objects stored into the catenating
  lists. */ 
typedef struct {
  int size;             //!< hash table size
  int card;             //!< number of stored items
  MyHash_nodeptr_array table;   //!< pointer to the hash table
} MyHash_table;

/* Hashtable Functions */
void MyHashtable_init(MyHash_table *ht, int n);
void MyHashtable_clear(MyHash_table *ht);
int MyHashtable_func(const char *s, int len, const MyHash_table *ht);
MyHash_node *MyHashtable_search(const char *s, int slen,
				const MyHash_table *ht);
int MyHashtable_insert(const char *s, int slen, int npattern, MyHash_table *ht
#ifdef DEBUG
		       , char *word, int len
#endif
		       );

MyHash_node *
CGrep_CheckIfIsPattern(const char *cbody, int cw_len, const MyHash_table *ht,
		       MyHash_node *hn, const char filter[]);


int CGrep_GetMatchingCW(MyHash_table *ht, char *filter, char **options,
			int npattern, const Console *c);


int * CGrep_GetCWOccurrences(int *, const MyHash_table *, const char *filter,
			     const char * cbody, size_t remaining);

/*
int CGrep_GetSeparators(MyHash_table *ht,
			char *filter, Dictionary *d, Console *c);
*/

proximity_hit_t *
CGrep_GetOccurrencesProximitySpaceless(int *, int, int,
				       const MyHash_table *,
				       const char filter[],
				       const MyHash_table *, const char *,
				       const char *, size_t);

proximity_hit_t *
CGrep_GetOccurrencesProximity(int *nocc, int prox_window,
			      int npatterns, const MyHash_table *ht,
			      const char filter[],
			      const MyHash_table *separators,
			      const char *sepFilter,
			      const char *buf, size_t buf_len,
			      const Hash_node *nl);

int * CGrep_SearchPattern(int *, const char *, size_t, const char *, char **);

int * CGrep_SearchWord(int *nres, const char *word, const char *ctext,
		       size_t ctext_len, int errors);

int * CGrep_SearchSubstring(int *nres, const char *pattern, const char *ctext,
			    size_t ctext_len, int errors);

proximity_hit_t *
CGrep_SearchProximity(int *nres, const char *ctext, size_t ctext_sz,
		      int prox_window, char **patterns,	char ***options);

const char *
CGrep_GetNextCWOccurrence(int *len, MyHash_table *ht, const char *filter,
			  const char *cbody, size_t remaining);

char * CGrep_escapeStringConfigurable(const char *s, size_t len, char min,
				      char max, const char *exceptions);

char * CGrep_escapeString(const char *s, size_t len);

