/** \file cgrep.c
 *
 * \brief Implementation of an agrep-like search program over Huffword
 *        compressed files, and with additional capabilities such as
 *        proximity search
 *
 * \author Paolo Ferragina and Alessandro Tommasi, Dipartimento di
 *           Informatica, Pisa (Italy)
 *    
 * \date March 2003
 *
 * Released under LGPL (see LICENSE file for license terms)
 */

#include "CGrepLib.h"
#include <fcntl.h>

/* option values, global */
int count = 0;
int window = 5;
int escape = 0;
int proximity = 0;
int prox_window = 0;
int lookup = 0;
int highlight = 0;
char *preHighlight = NULL;
char *postHighlight = NULL;
char *separator = NULL;
/* option values end */


#define DECOMPRESS_PREVIOUS(a, b, c, d, e, f, g, h) \
  if ( a ) \
     HDecompress_previousBlock_tokens_spaceless(b, c, d, e, f, g, h); \
  else \
     HDecompress_previousBlock_tokens(b, c, d, e, f, g, h); \

#define DECOMPRESS_NEXT(a, b, c, d, e, f, g, h) \
  if ( a ) \
     HDecompress_nextBlock_tokens_spaceless(b, c, d, e, f, g, h); \
  else \
     HDecompress_nextBlock_tokens(b, c, d, e, f, g, h); \


void usage(char *exname)
{
  printf("usage: %s [cgrep options] [--] [[agrep options] pattern]+ fname\n",
	 exname);
  puts("Search within filename (huffword-compressed file) for matches.");
  puts("CGREP OPTIONS:");
  puts("  -b  : highlight matching words in bold");
  puts("  -c  : count matches");
  puts("  -l  : lookup in the dictionary only (cannot be used with -w or -p)");
  puts("  -m pre post: prepend 'pre' and append 'post' to the matching words");
  puts("  -p n: search patterns within a proximity window of width n (n>0)");
  puts("  -s sep: output in format: match1\\nsep\\nmatch2... (not with -l)");
  puts("  -w n: return n words on left/right of the match (default 5, n>=0)");
  puts("  -x  : escape non-printable chars");
  puts("See 'man agrep' for agrep options.");
  puts("Options -n, -c, -s, -t, -G, -l to agrep are NOT allowed.");
}

char *
mapFile(size_t *input_sz, const char *fname, int canBeEmpty)
{
  struct stat sbuf;
  char *input_buf;
  FILE *input_file;

  stat(fname,&sbuf);
  if ( (input_file = fopen(fname, "r")) == NULL ) {
    fprintf(stderr, "cannot open %s, exiting.\n", fname);
    exit(2);
  }
  *input_sz = sbuf.st_size;
  
  if ( !canBeEmpty && (input_sz == 0) ) {
    fprintf(stderr,"Fatal Error: Input file empty\n");
    exit(2); 
  }

  input_buf = (char *) mmap(NULL,*input_sz,PROT_READ,MAP_SHARED,
			    fileno(input_file),0);
  if ( input_buf == MAP_FAILED ) {
    perror("mmap");
    exit(2);
  }
  fclose(input_file);
  return(input_buf);
}

int
parseCGrepOptions(int noptions, char **options)
{
  int i = 1;
  for (; i < noptions; i++) {
    if ( *options[i] != '-' ) // not an option
      return(i);
    switch (options[i][1]) {
    case 'c':
      count = 1; break;
    case 'b':
      highlight = 1;
      preHighlight = "[1m";
      postHighlight = "[0m";
      break;
    case 'm':
      highlight = 1;
      if ( (options[i+1] && options[i+2]) && (noptions > i+3) ) {
	preHighlight = options[i+1];
	postHighlight = options[i+2];
	i += 2;
      } else {
	usage(options[0]);
	fprintf(stderr, "Error: '-m' requires two arguments\n");
	exit(2);
      }
      break;
    case 'x':
      escape = 1; break;
    case 'w':
      if ( options[i][2] )
	window = atoi(options[i]+2);
      else if ( i < noptions - 1 )
	window = atoi(options[++i]);
      else {
	usage(options[0]);
	fprintf(stderr, "Error: '-w' requires an argument\n");
	exit(2);
      }
      if ( window < 0 ) {
	usage(options[0]);
	fprintf(stderr, "Error: invalid value for '-w' (%d)\n", window);
	exit(2);
      }
      break;
    case 'l':
      lookup = 1;
      if ( proximity ) {
	usage(options[0]);
	fprintf(stderr, "You cannot perform a lookup by proximity\n");
	exit(2);
      }
      break;
    case 's':
      separator = options[++i];
      break;
    case 'p':
      proximity = 1;
      if ( lookup ) {
	usage(options[0]);
	fprintf(stderr, "You cannot perform a lookup by proximity\n");
	exit(2);
      }
      if ( options[i][2] )
	prox_window = atoi(options[i]+2);
      else if ( i < noptions - 1 )
	prox_window = atoi(options[++i]);
      else {
	usage(options[0]);
	fprintf(stderr, "Error: '-p' requires the width of the proximity "
		"window\n");
	exit(2);
      }
      if ( !prox_window ) {
	usage(options[0]);
	fprintf(stderr, "Error: invalid width (0) for the proximity window\n");
	exit(2);
      }
      break;
    case '-':
      if ( !options[i][2] )
	return(i+1);
    default:
      usage(options[0]);
      fprintf(stderr, "Error: unrecognized option %s\n", options[i]);
      exit(2);
    }
  }
  usage(options[0]);
  exit(2);
  return(0); // just to avoid compiler warning
}

void CGrep_Lookup(const char *inf, const char *outf, char **options);

int do_lookup(char *input_buf, size_t input_sz, char *pattern,
	      char **options)
{
  char *myOptions[CGREP_MAX_AGREPOPTIONS];
  int i, j;
  char ifname[100];
  char ofname[100];
  char pids[10];
  char *buf;
  int len;
  int written = 0;
  int total = 0;
  int attempt = 0;
  int tmpfd;
  char *body;
  int blen;
  int tlen;
  Console console;
  Dictionary *d;

  sprintf(pids, "%d", getpid());

  HDecompress_getBodyAndConsole(input_buf, input_sz, &tlen, &body, &blen,
				&console);
  
  d = &(console.dictionary);

  do
    sprintf(ifname, "/tmp/agrep.dict.%s.%d", pids, ++attempt);
  while ( ((tmpfd = open(ifname, O_RDWR|O_CREAT|O_EXCL,
			 S_IREAD|S_IWUSR)) == -1) &&
	  (errno == EEXIST) );
    
  if ( tmpfd == -1 ) {
    perror("open");
    exit(2);
  }

  do {
    written =  write(tmpfd, d->content + written, d->length - written);
    if ( written == -1 ) {
      perror("write");
      exit(2);
    }
  } while ( written < d->length );
  close(tmpfd);


  sprintf(ofname, "/tmp/agrep.match.%s.%d", pids, attempt);

  myOptions[0] = CGREP_AGREP_EXECUTABLE;
  i = 1; j = 0;

  for ( ; options[j] && i < CGREP_MAX_AGREPOPTIONS; )
    myOptions[i++] = options[j++];
  myOptions[i++] = pattern;
  myOptions[i] = NULL;

  CGrep_Lookup(ifname, ofname, myOptions);

  unlink(ifname);

  buf = mapFile(&len, ofname, 1);

  written = 0;
  
  if ( count ) {
    if ( !buf ) puts("0");
    else {
      char *run = buf;
      int cnt = 0;
      while ( run < buf+len ) {
	if ( *run == '\n' ) cnt++;
	if ( (run > buf) && (run < buf+len-1) &&
	     ((*run) == '\n') && (*(run-1) == '\n') &&
	     (*(run+1) == '\n') ) cnt--;
	run++;
      }
      printf("%d\n", cnt);
    }
  } else if ( buf )
    while ( (total + (written = write(1, buf+total, len-total))) < len )
      total += written;

  if ( buf ) {
    munmap(buf, len);
    unlink(ofname);
    return(0);
  }
  unlink(ofname);
  return(1);
}

int isMatchingToken(int tr, proximity_hit_t ph, int npatterns)
{
  int i = 0;
  for ( ; i < npatterns; i++ )
    if ( ph.ranks[i] == tr ) return 1;
  return 0;
}

int main(int argc, char **argv)
{
  size_t input_sz;
  char *input_buf;
  int beginningAgrepOptions = -1;
  char ***options = (char ***) malloc(sizeof(char **) * argc);
  char **patterns = (char **) malloc(sizeof(char *) * argc);
  int i = 0;
  int npatterns;
  int *positions;
  int nresults = 0;
  proximity_hit_t *hits;

  if ( argc < 3 ) { /* at least a pattern and a file */
    usage(argv[0]);
    exit(2);
  }

  input_buf = mapFile(&input_sz, argv[argc-1], 0);

  beginningAgrepOptions = parseCGrepOptions(argc-1, argv);

  npatterns = 0;
  do {
    i = 0;
    options[npatterns] = (char **) malloc(sizeof(char *) * argc);
    while ( (*argv[beginningAgrepOptions] == '-') &&
	    (beginningAgrepOptions < argc - 1) ) {
      if ( !strcmp("-n", argv[beginningAgrepOptions]) ||
	   !strcmp("-c", argv[beginningAgrepOptions]) ) {
	usage(argv[0]);
	fprintf(stderr,
		"Error: option '%s' to agrep is not allowed\n",
		argv[beginningAgrepOptions]);
	exit(2);
      }
      options[npatterns][i++] = argv[beginningAgrepOptions++];
    }
    options[npatterns][i] = NULL;
    
    while ( (*argv[beginningAgrepOptions] != '-') &&
	    beginningAgrepOptions < argc - 1 ) {
      patterns[npatterns++] = argv[beginningAgrepOptions++];
      options[npatterns] = NULL;
    }
  } while ( beginningAgrepOptions < argc - 1 );

  patterns[npatterns] = NULL;
  options[npatterns] = NULL;
  if ( !npatterns ) {
    usage(argv[0]);
    fprintf(stderr, "Error: missing pattern\n");
    exit(2);
  }
  if ( (npatterns > 1) && ( !proximity ) ) {
    usage(argv[0]);
    fprintf(stderr, "Error: multiple patterns allowed only"
	    " for proximity search\n");
    exit(2);
  }

  if ( lookup ) {
    return(do_lookup(input_buf, input_sz, patterns[0], options[0]));
  }

  /* Check whether we are performing a proximity search */
  if ( proximity ) {
    hits = CGrep_SearchProximity(&nresults, input_buf, input_sz, prox_window,
				 patterns, options);
    if ( count )
      printf("%d\n", nresults);
    else if ( nresults ) {
      char *body;
      int tlen, blen;
      int scanned;
      Console console;
      HDecompress_getBodyAndConsole(input_buf, input_sz, &tlen, &body, &blen,
				    &console);
      for ( scanned = 0; scanned < nresults; scanned++ ) {
	int decoded, plen = 0;
	char *waste, *pstr;
	int tokens = 0;
	int currRank = hits[scanned].start_position;
	int currDispl = hits[scanned].byte_position;
	int needSpace = 0;

	if ( scanned && separator )
	  printf("%s\n", separator);

	DECOMPRESS_PREVIOUS(*input_buf, body+hits[scanned].byte_position,
			    hits[scanned].start_position,
			    window, &waste, &plen, &decoded, &console);
	if ( plen ) {
	  waste[plen] = '\0';
	  if ( escape ) {
	    pstr = CGrep_escapeString(waste, plen);
	    free(waste);
	  } else
	    pstr = waste;
	  printf(pstr);
	  if ( *input_buf && isalnum(pstr[strlen(pstr)-1])) {
	    // spaceless, add a space
	    putchar(' ');
	  }
	  fflush(stdout); /* otherwise it's not safe to free */
	  free(pstr);
	}
	while ( (unsigned int) tokens < hits[scanned].end_position -
    		hits[scanned].start_position + 1 ) {
	  int isMatching = isMatchingToken(currRank, hits[scanned], npatterns);
	  char *printed;

	  DECOMPRESS_NEXT(*input_buf, body+currDispl, input_sz-currDispl, 1,
			  &waste, &plen, &decoded, &console);
	  currDispl += decoded;
	  tokens++;
	  currRank++;
	  if ( needSpace && plen && isalnum(waste[0]) )
	    putchar(' ');
	  if ( plen ) {
	    waste[plen] = '\0';
	    if ( escape ) {
	      pstr = CGrep_escapeString(waste, plen);
	      free(waste);
	    } else
	      pstr = waste;
	    printed = pstr;
	    if ( isMatching && highlight && preHighlight ) {
	      // must make sure only the token is highlighted
	      char *end = pstr + strlen(pstr) - 1;
	      while ( (end >= pstr) && isalnum(*end) )
		end--;
	      while ( printed <= end )
		putchar(*(printed++));
	      printf(preHighlight);
	    }
	    printf(printed);
	    needSpace = ( *input_buf && isalnum(pstr[strlen(pstr)-1]));
	    fflush(stdout);
	    free(pstr);
	  }
	  if ( isMatching && highlight && postHighlight )
	    printf(postHighlight);
	}

	DECOMPRESS_NEXT(*input_buf, body+currDispl, input_sz-currDispl,
			window, &waste, &plen, &decoded, &console);
	if ( needSpace && plen && isalnum(waste[0]) )
	  putchar(' ');
	if ( plen ) {
	  waste[plen] = '\0';
	  if ( escape ) {
	    pstr = CGrep_escapeString(waste, plen);
	    free(waste);
	  } else
	    pstr = waste;
	  printf(pstr);
	  fflush(stdout);
	  free(pstr);
	}
	puts("");
      }
      if ( hits )
	free(hits);
    }
  } else { /* "standard" search for one pattern */
    positions = CGrep_SearchPattern(&nresults, input_buf, input_sz,
				    patterns[0], options[0]);
    if ( count ) {
      printf("%d\n", nresults);
    } else if ( nresults ) {
      Console console;
      char *body;
      int tlen, blen;
      int scanned;
      HDecompress_getBodyAndConsole(input_buf, input_sz, &tlen, &body, &blen,
				    &console);
      for ( scanned = 0; scanned < nresults; scanned++ ) {
	int decoded, plen = 0;
	char *waste, *pstr;
	int addSpace = 0;

	if ( scanned && separator )
	  printf("%s\n", separator);

	DECOMPRESS_PREVIOUS(*input_buf, body+positions[scanned],
			    positions[scanned], window,
			    &waste, &plen, &decoded, &console);
	if ( plen ) {
	  waste[plen] = '\0';
	  if ( escape ) {
	    pstr = CGrep_escapeString(waste, plen);
	    free(waste);
	  } else
	    pstr = waste;	     
	  printf(pstr);
	  if ( *input_buf && isalnum(pstr[strlen(pstr)-1])) {
	    // spaceless, add a space
	    putchar(' ');
	  }
	  fflush(stdout); /* otherwise it's not safe to free */
	  free(pstr);
	}
	if ( highlight && preHighlight )
	  printf(preHighlight);
	/* now, decompress the matching token */
	DECOMPRESS_NEXT(*input_buf, body+positions[scanned],
			input_sz - positions[scanned], 1,
			&waste, &plen, &decoded, &console);
	if ( plen ) {
	  waste[plen] = '\0';
	  if ( escape ) {
	    pstr = CGrep_escapeString(waste, plen);
	    free(waste);
	  } else
	    pstr = waste;
	  printf(pstr);
	  if ( highlight && postHighlight )
	    printf(postHighlight);
	  if ( *input_buf && isalnum(pstr[strlen(pstr)-1])) {
	    // spaceless, add a space
	    addSpace = 1;
	  }
	  fflush(stdout);
	  free(pstr);
	}
	DECOMPRESS_NEXT(*input_buf, body+positions[scanned]+decoded,
			input_sz - positions[scanned] - decoded, window,
			&waste, &plen, &decoded, &console);
	if ( plen ) {
	  waste[plen] = '\0';
	  if ( escape ) {
	    pstr = CGrep_escapeString(waste, plen);
	    free(waste);
	  } else
	    pstr = waste;
	  if ( addSpace && *input_buf && isalnum(pstr[0])) {
	    // spaceless, add a space
	    putchar(' ');
	  }
	  printf(pstr);
	  fflush(stdout);
	  free(pstr);
	}
	puts("");
      }
    }
    if ( positions )
      free(positions);
  }

  return((nresults>0)?0:1);
}
