/**
 * \file searchword.c
 *
 * \brief This file shows how to search for a word over a compressed
 *        file.
 *
 * \author Paolo Ferragina and Alessandro Tommasi, Dipartimento di
 *           Informatica, Pisa (Italy)
 *
 * \date March 2003
 *
 * Released under LGPL (see LICENSE file for license terms)
 */

#include "CGrepLib.h"

#define ERRORS 0

void usage()
{
  puts("usage: searchword word file.hwz");
}

int main(int argc, char **argv)
{
  struct stat sbuf;
  int input_sz, scanned = 0;
  FILE *input_file;
  char *input_buf;
  int *position = NULL;
  int nocc = 0;
  char *body;
  int tlen, blen;
  Console console;

  if ( argc < 2 ) { /* check for at least 2 arguments */
    usage();
    exit(0);
  }

  /* check for file readability */
  stat(argv[argc-1],&sbuf);
  if ( (input_file = fopen(argv[argc-1], "r")) == NULL ) {
    fprintf(stderr, "cannot open %s, exiting.\n", argv[argc-1]);
    exit(-1);
  }
  input_sz = (int)sbuf.st_size;
  if (input_sz == 0){ /* check for file size */
    fprintf(stderr,"Fatal Error: Input file empty\n");
    exit(-1); 
  } 

  /* mmap the compressed file */
  input_buf = (char *) mmap(NULL,input_sz,PROT_READ,MAP_SHARED,
			    fileno(input_file),0);
  if ( input_buf == MAP_FAILED ) {
    perror("mmap");
    exit(-1);
  }

  printf("searching for \"%s\" allowing %d errors...\n", argv[argc-2],
	 ERRORS);

  /* invoke CGrep_SearchWord. The function returns an array of positions
     of matching codeword within the body of the compressed file. The
     last parameters is the number of errors allowed in the search.
     (defined at the top of this file)
  */
  position = CGrep_SearchWord(&nocc, argv[argc-2], input_buf, input_sz,
			      ERRORS);
  
  /* check if some results were found */
  if ( position == NULL ) {
    puts("No matches found.");
    return(0);
  }
  /* otherwise, go ahead and decompress the compressed file around
     each match */
  /* Get the compressed body and console from the compressed file */
  HDecompress_getBodyAndConsole(input_buf, input_sz, &tlen, &body, &blen,
				&console);
  for ( scanned = 0; scanned < nocc; scanned++ ) { /* for each match */
    char *prev = NULL, *next = NULL;
    int prevLen = 0, nextLen = 0, decoded;
    printf("at position %d: ", position[scanned]);
    /* decompress 3 words to the left of the match */
    /* the first byte of the compressed file is != 0 if the file is compressed
       without spaces, 0 otherwise */
    if ( *input_buf ) { /* spaceless or with spaces? */
      HDecompress_previousBlock_tokens_spaceless(body+position[scanned],
						 position[scanned],
						 3, &prev, &prevLen, &decoded,
						 &console);
    } else {
      HDecompress_previousBlock_tokens(body+position[scanned],
				       position[scanned],
				       3, &prev, &prevLen, &decoded,
				       &console);
    }
    /* print the previous tokens */
    if ( prevLen ) {
      prev[prevLen] = '\0'; /* null-terminate the string */
      printf(prev);
      if ( *input_buf && isalnum(prev[prevLen-1]) ) { // add a space
	putchar(' ');
      }
    }
    /* now, decompress 4 words beginning with the match */
    /* the first byte of the compressed file is != 0 if the file is compressed
       without spaces, 0 otherwise */
    if ( *input_buf ) { /* spaceless or with spaces? */
      HDecompress_nextBlock_tokens_spaceless(body+position[scanned],
					     input_sz-position[scanned],
					     4, &next, &nextLen, &decoded,
					     &console);
    } else {
      HDecompress_nextBlock_tokens(body+position[scanned],
				   input_sz-position[scanned],
				   3, &next, &nextLen, &decoded,
				   &console);
    }
    /* print the match and the next 3 tokens */
    if ( nextLen ) {
      next[nextLen] = '\0'; /* null-terminate the string */
      puts(next);
    }
    fflush(stdout); /* flush the buffers */
    if ( prev ) free(prev); /* frees the buffers */
    if ( next ) free(next);
  }

  printf("(%d occurrences)\n", nocc);
  
  return(0);
}
