/*! \file HuffwordLib.c

  \brief Library of C procedures for canonical huffword
  (de)compression and its manipulation. The Library exports a set of C
  functions that allow to access and process all of its constitutent
  parts: dictionary, codewords, tokens, etc. etc.

  \author  Paolo Ferragina, Dipartimento di Informatica, Pisa (Italy).

  \date Version 1.0, March 2003. 

  This file is licensed under LGPL terms (see file LICENSE)
*/




/* -------------------------------------------------------- */
/* ---------------  Pull of data structures --------------- */
/* -------------------------------------------------------- */


#include "HuffwordLib.h"







/* -------------------------------------------------------------- */
/* -------------------- High-level procedures ------------------- */
/* --------------------                       ------------------- */
/* --------------------  Text (de)compression ------------------- */
/* -------------------------------------------------------------- */


/** Compresses the text string by using Compose_compressed_string()
 * and setting the first (reserved) byte of the produced string to 0,
 * denoting the case of 'plain compression'.
 *
 * @param text pointer to the text string to be compressed.
 * @param tlen length of the text string.
 * @param jump_value average char distance between consecutive jumpers, if 0 they are not computed.
 * @param cstring address of the pointer that will refer to the computed compressed string.
 * @param clen pointer to the integer that will contain the length of the compressed string.
 * @param Verbose if 1 the procedure prints infos on the compression process.
*/

void Huffw_compress(char *text, int tlen, int jump_value, char **cstring, int *clen, int Verbose)
{

  char *body;
  int blen;
  Console c;


  if ((tlen <= 0) || (text == NULL)){
    fprintf(stderr,"Fatal Error: You're trying to compress an empty file! (Huffw_compress)\n");
    exit(-1); }
    

  // Initializes the hastable then used by HParse_text()
  HHashtable_init(&(c.hashtable),tlen/1000+5);
  
  HParse_text(text,tlen,&(c.hashtable));  

  HCompress_getBodyAndConsole(text,tlen,jump_value,&body,&blen,&c);

  HCompress_getString(cstring,clen,tlen,body,blen,&c);

  free(body);

  // The first byte is set to 0 for the Classic model
  **cstring = 0;

  if(Verbose)
    Huffw_PrintInfo(&c,blen,tlen,*clen,0,Verbose); // Parsing rule = 0

}


/** Decompresses the compressed string; if the first (reserved) byte
 * is not 0, indicating not plain compression, an error is returned.
 *
 * @param cstring pointer to the compressed string.
 * @param clen length of the compressed string.
 * @param text address of the pointer to the uncompressed text string, to be computed.
 * @param tlen pointer to the integer that will contain the length of the uncompressed text.
 * @param Verbose if 1 the procedure prints infos on the compressed string.
*/

void Huffw_decompress(char *cstring, int clen, char **text, int *tlen, int Verbose)
{
  Console c;
  char *body;
  int blen;

  if ((clen <= 0) || (cstring == NULL)){
    fprintf(stderr,"Fatal Error: You're trying to decompress an empty file! (Huffw_decompress)\n");
    exit(-1); }

  // Checks if the file has been compressed with Classic model
  if(*cstring != 0){
    fprintf(stderr,"Fatal Error: First byte of compressed file not 0, actually %d\n", *cstring);
    exit(-1); }


  HDecompress_getBodyAndConsole(cstring,clen,tlen,&body,&blen,&c);

  HDecompress_nextBlock_bytes(body,blen,text,tlen, &c);

  if(Verbose)
    Huffw_PrintInfo(&c,blen,*tlen,clen,0,Verbose); // Parsing rule = 0

}




/** Compresses the text string according to the Spaceless model. It
 * uses Compose_compressed_string() to put together all the infos and
 * the Spaceless body. It sets the first (reserved) byte to 1,
 * denoting the case of 'spaceless compression'.
 *
 * @param text pointer to the text string to be compressed.
 * @param tlen length of the text string.
 * @param cstring address of the pointer that will refer to the computed compressed string.
 * @param clen pointer to the integer that will contain the length of the compressed string.
 * @param Verbose if 1 the procedure prints infos on the compression process.
*/


void Huffw_spaceless_compress(char *text, int tlen, char **cstring, int *clen, int Verbose)
{

  Console c;
  char *body, *body_spaceless;
  int blen, blen_spaceless;
  int jump_value = 0; // it is forced to be 0

  if ((tlen <= 0) || (text == NULL)){
    fprintf(stderr,"Fatal Error: You're trying to compress an empty file! (Huffw_spaceless_compress)\n");
    exit(-1); }

  // Initializes the hastable then used by HParse_text()
  HHashtable_init(&(c.hashtable),tlen/1000+5);
  
  HParse_text(text,tlen,&(c.hashtable));  

  HCompress_getBodyAndConsole(text,tlen,jump_value,&body,&blen,&c);

  HCompress_contractSpaces(body,blen,&body_spaceless, &blen_spaceless, &c);
  
  free(body);

  HCompress_getString(cstring,clen,tlen,body_spaceless,blen_spaceless,&c);

  free(body_spaceless);

  // The first reserved byte is set to 1 for the Spaceless model 
  **cstring = 1;

  if(Verbose)
    Huffw_PrintInfo(&c,blen_spaceless,tlen,*clen,1,Verbose); // Parsing rule = 1

}




/** Decompresses the compressed string produced according to the
 * Spaceless model. If the the first (reserved) byte is not 1, indicating 
 * not spaceless compression, an error is returned.
 *
 * @param cstring pointer to the compressed string.
 * @param clen length of the compressed string.
 * @param text address of the pointer to the uncompressed text string, to be computed.
 * @param tlen pointer to the integer that will contain the length of the uncompressed text.
 * @param Verbose if 1 the procedure prints infos on the compressed string.
*/

void Huffw_spaceless_decompress(char *cstring, int clen, char **text, int *tlen, int Verbose)
{
  Console c;
  char *body, *body_spaceless;
  int blen, blen_spaceless;

  if ((clen <= 0) || (cstring == NULL)){
    fprintf(stderr,"Fatal Error: You're trying to decompress an empty file! (Huffw_spaceless_decompress)\n");
    exit(-1); }


  // Checks if the file has been compressed with the Spaceless model
  if(*cstring != 1){
    fprintf(stderr,"Fatal Error: First byte of compressed file not 1, actually %d\n", *cstring);
    exit(-1); }

  // the jump_value is set to be 0 for the Spaceless model 
  HDecompress_getBodyAndConsole(cstring,clen,tlen,&body_spaceless,&blen_spaceless,&c);

  HDecompress_expandSpaces(body_spaceless, blen_spaceless, &body, &blen,&c);
  
  HDecompress_nextBlock_bytes(body,blen,text,tlen,&c);

  if(Verbose)
    Huffw_PrintInfo(&c,blen_spaceless,*tlen,clen,1,Verbose); // Parsing rule = 1

}



/** Print compression infos. It is offsetcw[] > num_tokens if the
 *   corresponding cwlen is not occurring. If Verbose > 1 also the
 *   dictionary tokens and their codewords are printed.
 *
 * @param c pointer to the Console data structure containing all the compression infos.
 * @param ctext_len length of the compressed text.
 * @param text_len length of the plain text.
 * @param cstring_len length of the overall compressed file (including header, dictionary, and compressed body).
 * @param rule indicates the parsing rule, 0 = Plain, 1 = Spaceless.
 * @param Verbose set to 1 for printing out also the dictionary tokens and their codewords.
 */

void Huffw_PrintInfo(Console *c, int ctext_len, int text_len, int cstring_len, int rule, int Verbose)
{
  int i,j;
  float f = ((float)cstring_len)/((float)text_len);

  if (c->jumpers.number){
    j = text_len/c->jumpers.number;
  } else { j=0; }
  printf("\n------------------- Global infos --------------------\n\n");
  printf("Parsing rule = %d\n", rule);
  printf("Text length = %d\n", text_len);
  printf("Compressed body length = %d\n", ctext_len);
  printf("Dictionary size = %d, Dictionary length = %d\n",c->dictionary.num_tokens,c->dictionary.length);
  printf("Max codeword length = %d\n",c->canoinfo.max_cwlen);
  printf("Number of jumpers = %d, average jump size = %d chars\n\n", c->jumpers.number,j);
  for(i=c->canoinfo.max_cwlen; i>0;i--)
    printf("Offset[%d] = %d, First_cw[%d] = %x\n",i,c->canoinfo.offsetcw[i],i,c->canoinfo.firstcw[i]);
  printf("\n Overall compressed file length = %d, compression ratio (percentage) = %.1f \n\n", 
	 cstring_len, f * 100);

  if (Verbose > 1)
    HDictionary_print(&(c->dictionary), &(c->hashtable), 1);

}



/* ------------------------------------------------------ */
/* --------------------  Text parsing ------------------- */
/* ------------------------------------------------------ */



/** Parses the text using the procedure HToken_getLengthNext() and
 * fills in the hash table with items containing for each token, its
 * number of occurrences in the text (count_occ), the token string
 * (str) and its length (len_str). These infos will be then used by
 * the procedure to construct the Huffman tree. The programmer has to
 * take care of the initialization of the hash table, via
 * HHashtable_init(). This way, it is possible to parse a sequence of
 * texts and build one unique Huffamn tree for them.
 *
 * @param text pointer to the string to be parsed.
 * @param text_len length of the string to be parsed.
 * @param ht pointer to the hash table that will contain the parsed tokens and their countings.
 */


void HParse_text(char *text, int text_len, HHash_table *ht)
{
  int token_len,i;
  char *token;


  if ((text_len <= 0) || (text == NULL)){
    fprintf(stderr,"Fatal Error: You're trying to parse an empty file! (HParse_text)\n");
    exit(-1); }

  for(i=0; i<text_len; i+=token_len){

    // get the next token from the input
    HToken_getLengthNext(text+i,text_len-i,&token_len);
    token = text+i;

    // Insert token into hash table and count if it is new
    HHashtable_insert(token,token_len,ht);   
  }

}




/* -------------------------------------------------------------- */
/* --------------------  Refined procedures   ------------------- */
/* --------------------                       ------------------- */
/* --------------------  Text (de)compression ------------------- */
/* -------------------------------------------------------------- */



/** Construct the byte-aligned and tagged Huffword string, exploiting
 * the tagged and byte-aligned codewords available in the hash-table
 * data structure. 
 *
 * @param text pointer to the string to be compressed.
 * @param text_len length of the string to be compressed.
 * @param ht pointer ot the hash table containing the coded tokens and their codewords.
 * @param ctext address of the string pointer that will refer to the compressed string.
 * @param ctext_len address of the integer that will contain the length of the compressed string.
 */

void HCompress_getBody(char *text, int text_len, HHash_table *ht, char **ctext, int *ctext_len)

{
  int i;
  int token_len, ctext_allocated;
  char *token;
  int taggedcw;
  int taggedcwlen;
  char *s;
  float resize_factor = 1.4;  // 40% more space


  if ((text_len <= 0) || (text == NULL)){
    fprintf(stderr,"Fatal Error: You're trying to compress an empty file! (HCompress_getBody)\n");
    exit(-1); }


  *ctext_len=0;
  ctext_allocated = text_len/2;
  *ctext = (char *) malloc((ctext_allocated+1) * sizeof(char));

  if((*ctext == NULL) || (text_len == 0)){
    fprintf(stderr,"Error: memory allocation (HCompress_getBody)\n");
    exit(-1);  }

  for(i=0; i<text_len; i+=token_len){
    
    // get the next token from the input
    token = text+i;
    HToken_getLengthNext(text+i,text_len-i,&token_len);

    // compute its tagged codeword
    HCodeword_TaggedFromToken(token,token_len,ht,&taggedcw,&taggedcwlen);

    // Writes the codeword into the body string
    HCodeword_tostring(taggedcw,taggedcwlen,&s);

    // Resize the space allocated to the compressed text, if needed
    if( (*ctext_len+taggedcwlen) >= ctext_allocated ) {
      *ctext = (char*) realloc(*ctext,(resize_factor*ctext_allocated + taggedcwlen + 1)*sizeof(char));
      if ( !(*ctext) ) { fprintf(stderr, "memory exhausted in HCompress_getBody\n"); exit(-1); }
      ctext_allocated = ctext_allocated * resize_factor + taggedcwlen + 1;
    }

    // Writes the tagged codeword into the compressed text
    memcpy((*ctext) + (*ctext_len),s,taggedcwlen);
    (*ctext_len) += taggedcwlen;

    free(s); // free the space allocated for the codeword
  }
}





/** Builds all the data structures needed for the Huffword
 * compression. Actually it builds the Huffman tree, and then computes
 * the Canonical data structure, the Dictionary and derives the
 * compressed body string as well the Jumpers, if jump_value is
 * greater than zero. 
 *
 * @param text pointer to the text string to be compressed.
 * @param tlen text length.
 * @param jump_value average number of chars to skip between consecutive jumpers (if 0, no jumpers).
 * @param body address of the pointer that will contain the compressed body string (mmapped from ctext).
 * @param blen pointer to the integer that will contain the length of the compressed body.
 * @param console pointer to the data structure containing all the compression info.
 */

void HCompress_getBodyAndConsole(char *text, int tlen, int jump_value, char **body, int *blen, 
				Console *console)
{

  Hash_nodeptr_array tree; 
  Canonical *cano = &(console->canoinfo);
  Dictionary *dict = &(console->dictionary);
  Jumpers *jump = &(console->jumpers);
  HHash_table *ht = &(console->hashtable);


  if ((tlen <= 0) || (text == NULL)){
    fprintf(stderr,"Fatal Error: You're trying to compress an empty file! (HCompress_getBodyAndConsole)\n");
    exit(-1); }


  // computes the huffman tree
  Hufftree_build(&tree,ht); 

  // computes the canonical info and updates the hashtable determining the (plain) codewords
  HCanonical_fromtree(tree,cano,ht); 

  // computes the dictionary data structure
  HDictionary_fromtree(tree,cano->num_tokens,dict); 

  // computes the body of the compressed text 
  HCompress_getBody(text,tlen,ht,body,blen);
   
  // computes the jumpers according to the jump_value specified, if 0 nothing is done
  HJumpers_fromds(text,tlen,ht,jump_value,jump);

}




/** Composes the entire compressed string allocating it from scratch.
 * It will be pointed to by *ctext and of length *ctext_len. Recall
 * that offsetcw[i] is set ideally to +infty if no codeword of length
 * 'i' is present in the dictionary. The composed string is formed as
 * follows:
 * - Compression type (1 byte), 0 indicates 'plain' and 1 'spaceless';
 * - Text length (4 bytes); 
 * - Number of tokens in the dictionary (4 bytes); 
 * - Max codeword length (4 bytes); 
 * - Firstcw[1], Offsetcw[1], ..., Firstcw[4], Offsetcw[4] (32 bytes); 
 * - Number of jumpers (4 bytes); 
 * - Array of text and ctext jumpers (8 bytes each pair), alternating [possibly absent];
 * - The dictionary length (4 bytes); 
 * - The dictionary content, as a string of tokens separated by 'newline';
 * - Compressed text length (4 bytes)
 * - Compressed text;
 *
 * @param ctext address of the pointer that will refer to the compressed file string, to be computed.
 * @param ctext_len pointer to the integer that will contain the length of the compressed file.
 * @param text_len length of the text to be compressed.
 * @param body pointer to the compressed text (already computed).
 * @param blen length of the compressed text.
 * @param console pointer to the data structure containing all the compression infos.
 *
 */

void HCompress_getString(char **ctext, int *ctext_len, int text_len,char *body, int blen, 
			Console *console)
{
  int global_length;
  Canonical *cano = &(console->canoinfo);
  Dictionary *dict = &(console->dictionary);
  Jumpers *jump = &(console->jumpers);


  if ((text_len <= 0) || (body == NULL) || (blen <=0)){
    fprintf(stderr,"Fatal Error: You're trying to compress an empty file! (HCompress_getString)\n");
    exit(-1); }


  global_length = 57 + 8 * jump->number +  dict->length + blen;

  *ctext = (char *) malloc(global_length * sizeof(char));
  *ctext_len=0;

  // reserved byte for storing compression rule
  *ctext_len += 1;

  HInt_tostring(*ctext + *ctext_len, text_len); 
  *ctext_len += 4;

  HInt_tostring(*ctext + *ctext_len, dict->num_tokens); 
  *ctext_len += 4;

  HCanonical_tostring(*ctext + *ctext_len, cano); 
  *ctext_len += 36;

  HJumpers_tostring(*ctext + *ctext_len, jump);
  *ctext_len += 8 * jump->number + 4;

  HInt_tostring(*ctext + *ctext_len, dict->length);  
  *ctext_len += 4;
  memcpy(*ctext + *ctext_len, dict->content, dict->length);
  *ctext_len += dict->length;

  HInt_tostring(*ctext + *ctext_len, blen);
  *ctext_len += 4;
  memcpy(*ctext + *ctext_len, body, blen);
  *ctext_len += blen;

  if (global_length != *ctext_len){
	fprintf(stderr,"Error: computing the compressed file length (HCompress_getString)\n");
	exit(-1);  }  


}


/** Removes the single spaces which occur between two consecutive
 *   alphanumeric tokens. This is the so called Spaceless model. 
 *
 * @param body pointer to the compressed body string.
 * @param blen length of the compressed body.
 * @param body_spaceless address of the pointer to the body compressed according to Spaceless model.
 * @param blen_spaceless length of the spaceless body string.
 * @param console pointer to the data structure containing all the compression info.
*/

void HCompress_contractSpaces(char *body, int blen, char **body_spaceless, int *blen_spaceless, Console *console)
{
  char *token, *sstring;
  int token_len, tcw_len, i;
  int spacecw, spacecw_len;
  char spacecw_string[4];
  int state, next_state;


  if ((body == NULL) || (blen <=0)){
    fprintf(stderr,"Fatal Error: You're trying to compress an empty file! (HCompress_contractSpaces)\n");
    exit(-1); }

  // Note that the dictionary contain the space if it has to be dropped
  if(HHashtable_search(" ",1,&(console->hashtable)) == NULL){ // No single space to be contracted
    *body_spaceless = body;
    *blen_spaceless = blen;
  } else {

    *body_spaceless = (char *) malloc(blen * sizeof(char));
    if (*body_spaceless == NULL) {
      fprintf(stderr,"Fatal Error: No memory for spaceless body (HCompress_contractSpaces\n");
      exit(-1); }

    // Get the codeword for the single space, put it into a string
    // Note that the dictionary contain the space if it has to be dropped
    HCodeword_TaggedFromToken(" ", 1, &(console->hashtable), &spacecw, &spacecw_len); 
    HInt_tostring(spacecw_string,spacecw);
    sstring = spacecw_string + (4-spacecw_len);

    *blen_spaceless = 0;

    // state of the computation: 
    // 0 = start
    // 1 = ALNUM token has been read, 
    // 2 = ALNUM token + space have been read
    // 3 = ALNUM  token + space +  ALNUM token have been read ==> drop the space.
    state = 0; 
    
    for(i=0; i<blen; i += tcw_len) {
      
      HToken_decompressNext(body+i,blen-i,&token,&token_len,&tcw_len,console);
      
      // Compute the next automaton status
      if ((state == 0) && (isalnum(*token)))
	next_state = 1;
      else if ((state == 1) && (memcmp(token," ",1) == 0) & (token_len == 1))
	next_state = 2;
      else if ((state == 1) && (isalnum(*token)))
	next_state = 1;
      else if ((state == 2) && (isalnum(*token)))
	next_state = 3;
      else next_state = 0;
      
      free(token);

      // Write the new body file
      if ((state == 2) && (next_state != 3)) { // suspended space is written out
	memcpy(*body_spaceless + *blen_spaceless, sstring, spacecw_len);
	*blen_spaceless += spacecw_len;
      }
      if (next_state != 2){ // no-singlespace token is written out
	memcpy(*body_spaceless + *blen_spaceless, body+i, tcw_len); 
	*blen_spaceless += tcw_len;
      }
      
      // Update the automaton statos
      if (next_state == 3) 
	next_state = 1;
      state = next_state;
      
    }

    // suspended space, at the end of the file, is written out
    if (state == 2) { 
      memcpy(*body_spaceless + *blen_spaceless, sstring, spacecw_len);
      *blen_spaceless += spacecw_len;
    }
  }

}



/** Initializes all data structures and the Console loading their
 * content from the compressed string (including the header,
 * dictionary, ...). Recall that it is offsetcw[i] > #tokens if does
 * not exist any codeword of length i. The hash table contains for
 * each token its corresponding tagged codeword and its length in
 * bytes. 
 * @param ctext pointer to the string that contains the compressed file string.
 * @param ctext_len length of the compressed file string.
 * @param text_len pointer to the integer that will contain the length of the uncompressed text.
 * @param body address of the pointer that will contain the compressed body string (mmapped from ctext).
 * @param blen pointer to the integer that will contain the length of the compressed body.
 * @param console pointer to the data structure containing all the compression info.
 */

void HDecompress_getBodyAndConsole(char *ctext, int ctext_len, int *text_len, char **body, int *blen, 
				  Console *console)
{
  int i;
  int len;
  Canonical *cano = &(console->canoinfo);
  Dictionary *dict = &(console->dictionary);
  Jumpers *jump = &(console->jumpers);
  HHash_table *hashtable = &(console->hashtable);



  if ((ctext == NULL) || (ctext_len <=0)){
    fprintf(stderr,"Fatal Error: You're trying to decompress an empty file! (HDecompress_getBodyAndConsole)\n");
    exit(-1); }


  i=0;

  // Reserved byte for storing compression rule
  i += 1;

  *text_len = HInt_fromstring(ctext + i); 
  i += 4;

  cano->num_tokens = HInt_fromstring(ctext + i); 
  i += 4;

  HCanonical_fromstring(ctext + i, cano); 
  i += 36;

  HJumpers_fromstring(ctext + i, jump);
  i += 8 * jump->number + 4;

  len = HInt_fromstring(ctext + i);  
  i += 4;
  HDictionary_fromstring(ctext+i,len,cano->num_tokens,dict);
  i += dict->length;

  HHashtable_fromdict(dict,cano,hashtable);

  *blen = HInt_fromstring(ctext + i);
  i += 4;
  *body =  ctext + i;
  i += *blen;

  if (i != ctext_len){
	fprintf(stderr,"Error: computing the compressed file length (HDecompress_getBodyAndConsole)\n");
	exit(-1);  }  

}




/** Re-insert the single spaces that were dropped during the Spaceless compression. The
 * programmer must take care to apply this procedure in the right case. 
 *
 * @param body_spaceless pointer to the Spaceless compressed body.
 * @param blen_spaceless length of the spaceless body string.
 * @param body address of the pointer that will contain the compressed body string (with spaces).
 * @param blen pointer to the integer that will contain the length of the compressed body.
 * @param console pointer to the data structure containing all the compression info.
*/

void HDecompress_expandSpaces(char *body_spaceless, int blen_spaceless, char **body, int *blen, Console *console)
{

  char *token, *sstring;
  int token_len, tcw_len, i;
  int spacecw, spacecw_len;
  char spacecw_string[4];
  int state, next_state;

  // To manage the dummy case
  if ( (body_spaceless == NULL) || (blen_spaceless <= 0) ) {
    *body = NULL;
    *blen = 0;
    return ;
  }

  // To manage a decompression point not aligned to a codeword
  *body = (char *) malloc(blen_spaceless * 2 * sizeof(char));
  *blen = 0;
  if (*body == NULL) {
    fprintf(stderr,"Fatal Error: No memory for body (HDecompress_expandSpaces\n");
    exit(-1); }

  // Get the codeword for the single space, put it into a string
  // Note that the dictionary contain the space if it has been dropped
  if(HHashtable_search(" ",1,&(console->hashtable)) == NULL){
    *body = body_spaceless;
    *blen = blen_spaceless;
  } else {
  
    HCodeword_TaggedFromToken(" ", 1, &(console->hashtable), &spacecw, &spacecw_len); 
    HInt_tostring(spacecw_string,spacecw);
    sstring = spacecw_string + (4-spacecw_len);

    // state of the computation: 
    // 0 = start
    // 1 = ALNUM token has been read, 
    // 3 = ALNUM token + ALNUM token have been read
    
    state = 0; 
    
    for(i=0; i<blen_spaceless; i += tcw_len) {
      
      HToken_decompressNext(body_spaceless+i,blen_spaceless-i,&token,&token_len,&tcw_len,console);
      
      // Compute the next automaton status
      if ((state == 0) && (isalnum(*token)))
	next_state = 1;
      else if ((state == 1) && (isalnum(*token)))
	next_state = 3;
      else next_state = 0;
      
      free(token);

      // Write the new body file
      if (next_state == 3) { // dropped space is written out
	memcpy(*body + *blen, sstring, spacecw_len);
	*blen += spacecw_len;
      }
      memcpy(*body + *blen, body_spaceless+i, tcw_len); 
      *blen += tcw_len;
      
      // Update the automaton statos
      if (next_state == 3)
	next_state = 1;
      state = next_state;
    }
  }
}




/** Decompresses a piece of compressed text for the specified number
 * of bytes. The Console data structure must be initialized. The programmer 
 * must ensure that the requested number of bytes do exist. 
 * It does not work well on the spaceless model (ie. first reserved byte is 1). 
 * In this case the decompressed string is without spaces between alphanumeric 
 * tokens. Here, the programmer must use the proper procedure.
 *
 * @param ctext pointer to the block to be decompressed.
 * @param num_bytes number of bytes to decompress.
 * @param text address of the pointer to the decompressed block (allocated by the procedure).
 * @param text_len address of the integer that will contain the length of the decompressed block.
 * @param console pointer to the data structure that contains all the compression infos.
 */

void HDecompress_nextBlock_bytes(char *ctext, int num_bytes, char **text, int *text_len, Console *console)
{
  int i;
  int token_len, allocated_text;
  char *token;
  int taggedcwlen;
  float resize_factor = 1.4; // 40% more

  // To manage the dummy case
  if ( (ctext == NULL) || (num_bytes <= 0) ) {
    *text = NULL;
    *text_len = 0;
    return ;
  }

  // To manage the case of decompression point not aligned to a codeword
  if (((*ctext) & (0x80)) == 0){
    fprintf(stderr,"Error: not a tagged byte (HDecompress_nextBlock_bytes)\n");
    exit(-1);  }

  *text_len=0;
  allocated_text = 2 * num_bytes;
  *text = (char *) malloc(allocated_text * sizeof(char));

  if((*text == NULL) || (num_bytes == 0)){
    fprintf(stderr,"Error: memory allocation (HDecompress_nextBlock_bytes)\n");
    exit(-1);  }

  for(i=0; i<num_bytes; i+=taggedcwlen){
    
    // get next token from the compressed text
    HToken_decompressNext(ctext+i,num_bytes-i,&token,&token_len,&taggedcwlen,console);

    // Resize the space allocated to the text, if needed
    if( (*text_len+token_len) >= allocated_text) {
      *text = (char*)realloc(*text,(resize_factor*allocated_text+1+token_len)*sizeof(char));
      if ( !(*text) ) { fprintf(stderr, "memory exhausted in HDecompress_nextBlock_bytes\n"); exit(-1); }
      allocated_text = allocated_text * resize_factor + 1 + token_len;
    }

    // Writes the token into the text
    memcpy((*text) + (*text_len),token,token_len);
    (*text_len) += token_len;

    free(token);
  }

}





/** Decompresses a piece of compressed text for the specified number
 * of bytes in the Spaceless model. The Console data structure must be
 * initialized. The programmer must ensure that the requested number
 * of bytes do exist and that we are in the case of the spaceless 
 * compression (ie. first reserved byte is 1).
 *
 * @param ctext pointer to the block to be decompressed.
 * @param num_bytes number of bytes to decompress.
 * @param text address of the pointer to the decompressed block (allocated by the procedure).
 * @param text_len address of the integer that will contain the length of the decompressed block.
 * @param console pointer to the data structure that contains all the compression infos.
 */

void HDecompress_nextBlock_bytes_spaceless(char *ctext, int num_bytes, char **text, 
					  int *text_len, Console *console)
{
  int exp_len;
  char *exp_text;
  

  // To manage the dummy case
  if ( (ctext == NULL) || (num_bytes <= 0) ) {
    *text = NULL;
    *text_len = 0;
    return ;
  }

  // To manage the case of decompression point not aligned to a codeword
  if (((*ctext) & (0x80)) == 0){
    fprintf(stderr,"Error: not a tagged byte (HDecompress_nextBlock_bytes_spaceless)\n");
    exit(-1);  }

  // Reinsert the spaces
  // Even if num_bytes is not alligned to a codeword, the procedure re-aligns it
   HDecompress_expandSpaces(ctext, num_bytes, &exp_text, &exp_len,console);

   // Decompresses enough text
   HDecompress_nextBlock_bytes(exp_text, exp_len, text, text_len, console);

}






/** Decompresses a piece of compressed text lying to the left of
 * "ctext" for a specified number of bytes. The Console data structure
 * must be initialized. The programmer must ensure that the requested number 
 * of bytes do exist.  
 * It does not work well on the spaceless model (ie. first reserved byte is 1). 
 * In this case the decompressed string is without spaces between 
 * alphanumeric tokens. Here, the programmer must use the proper procedure.
 *
 * @param ctext pointer to the block to be decompressed.
 * @param num_bytes number of bytes to decompress.
 * @param text address of the pointer to the decompressed block (allocated by the procedure).
 * @param text_len address of the integer that will contain the length of the decompressed block.
 * @param console pointer to the data structure that contains all the compression infos.
 */

void HDecompress_previousBlock_bytes(char *ctext, int num_bytes, char **text, int *text_len, 
				     Console *console)
{
  int i;
  int token_len, allocated_text;
  char *token;
  int taggedcwlen;
  float resize_factor = 1.4; // 40% more


  // To manage the dummy case
  if ( (ctext == NULL) || (num_bytes <= 0) ) {
    *text = NULL;
    *text_len = 0;
    return ;
  }

  // To manage the case of decompression point not aligned to a codeword
  if (((*ctext) & (0x80)) == 0){
    fprintf(stderr,"Error: not a tagged byte (HDecompress_previousBlock_bytes)\n");
    exit(-1);  }

  *text_len=0;
  allocated_text = 2 * num_bytes;
  *text = (char *) malloc(allocated_text * sizeof(char));

  if((*text == NULL) || (num_bytes == 0)){
    fprintf(stderr,"Error: memory allocation (HDecompress_previousBlock_bytes)\n");
    exit(-1);  }

  // compute the starting pos of the window
  for(i=0; i<num_bytes; i+=taggedcwlen){
    
    // get previous token from the compressed text
    HToken_decompressPrevious(ctext-i,num_bytes-i,&token,&token_len,&taggedcwlen,console);

    free(token);

  }

  i -= taggedcwlen;

  for(; i>=0; i-=taggedcwlen){
    
    // get next token from the compressed text
    HToken_decompressNext(ctext-i,num_bytes-i,&token,&token_len,&taggedcwlen,console);

    // Resize the space allocated to the text, if needed
    if( (*text_len+token_len) >= allocated_text) {
      *text = (char*)realloc(*text,(resize_factor*allocated_text+1+token_len)*sizeof(char));
      if ( !(*text) ) { fprintf(stderr, "memory exhausted in HDecompress_previousBlock_bytes\n"); exit(-1); }
      allocated_text = allocated_text * resize_factor + token_len + 1;
    }

    // Writes the token into the text
    memcpy((*text) + (*text_len),token,token_len);
    (*text_len) += token_len;
    
    free(token);
  }

}


/** Decompresses a piece of compressed text to the left of "ctext"
 * pointer for "num_bytes" bytes in the spaceless model. The Console
 * data structure must be initialized. The programmer must ensure that
 * the requested number of bytes do exist and that the string has been
 * compressed via 'spaceless model' (ie. first reserved byte is 1). 
 *
 * @param ctext pointer to the block to be decompressed.
 * @param num_bytes number of bytes to decompress.
 * @param text address of the pointer to the decompressed block (allocated by the procedure).
 * @param text_len address of the integer that will contain the length of the decompressed block.
 * @param console pointer to the data structure that contains all the compression infos.
 */

void HDecompress_previousBlock_bytes_spaceless(char *ctext, int num_bytes, char **text, 
					      int *text_len, Console *console)
{
  int exp_len;
  char *exp_text;
  

  // To manage the dummy case
  if ( (ctext == NULL) || (num_bytes <= 0) ) {
    *text = NULL;
    *text_len = 0;
    return ;
  }

  if (((*ctext) & (0x80)) == 0){
    fprintf(stderr,"Error: not a tagged byte (HDecompress_previousBlock_bytes_spaceless)\n");
    exit(-1);  }


  // Aligns to a tagged byte
  while( ( (*(ctext-num_bytes)) & (0x80) ) == 0){
    num_bytes++;
  }
   

  // Reinsert the spaces
   HDecompress_expandSpaces(ctext-num_bytes, num_bytes, &exp_text, &exp_len,console);

   // Decompresses enough text
   HDecompress_nextBlock_bytes(exp_text, exp_len, text, text_len, console);

}




/** Decompresses a piece of compressed text for a specified number of
 * alphanumeric tokens. The Console data structure must be
 * initialized. 
 * It does not work well on the spaceless model (ie. first reserved byte is 1). 
 * In this case  the decompressed string is without spaces between 
 * alphanumeric tokens. Here, the programmer must use the proper procedure.
 *
 * @param ctext pointer to the block to be decompressed.
 * @param bytes_left number of bytes yet available to be decompressed.
 * @param num_obj number of alphanumeric tokens to be decompressed.
 * @param text address of the pointer to the decompressed block (allocated by the procedure).
 * @param text_len address of the integer that will contain the length of the decompressed block.
 * @param decoded address of the integer indicating the number of decompressed bytes.
 * @param console pointer to the data structure that contains all the compression infos.
 */

void HDecompress_nextBlock_tokens(char *ctext, int bytes_left, int num_obj, 
				  char **text, int *text_len, int *decoded,
				  Console *console)
{
  int token_len;
  char *token;
  int taggedcwlen;


  // To manage the dummy case
  if ( (ctext == NULL) || (bytes_left <= 0) || (num_obj <= 0) ) {
    *text = NULL;
    *text_len = 0;
    *decoded = 0;
    return ;
  }

  // To manage a decompression point not aligned to a codeword
  if (((*ctext) & (0x80)) == 0){
    fprintf(stderr,"Error: not a tagged byte (HDecompress_nextBlock_tokens)\n");
    exit(-1);  }

  for(*decoded=0; (*decoded < bytes_left) && (num_obj > 0); *decoded += taggedcwlen){
    
    // get next token from the compressed text
    HToken_decompressNext(ctext + *decoded,bytes_left-*decoded,
			  &token,&token_len,&taggedcwlen,console);

    if (isalnum(token[0])){
      num_obj--;
    }

    free(token);
  }

  // Decompresses just the succeeding "i" bytes that contain the "num_object" words
  HDecompress_nextBlock_bytes(ctext,*decoded,text,text_len,console);
 
}




/** Decompresses a piece of compressed text for a specified number of
 * alphanumeric tokens in the spaceless model. The Console data
 * structure must be initialized. The programmer must ensure that the
 * compression type is 'spaceless' (ie. first reserved byte is 1). 
 *
 * @param ctext pointer to the block to be decompressed.
 * @param bytes_left number of bytes yet available to be decompressed.
 * @param num_obj number of alphanumeric tokens to be decompressed.
 * @param text address of the pointer to the decompressed block (allocated by the procedure).
 * @param text_len address of the integer that will contain the length of the decompressed block.
 * @param decoded address of the integer indicating the number of decompressed bytes.
 * @param console pointer to the data structure that contains all the compression infos.
 */

void HDecompress_nextBlock_tokens_spaceless(char *ctext, int bytes_left, int num_obj, 
					    char **text, int *text_len, int *decoded,
					    Console *console)
{
  int text_tmp_len,exp_len;
  char *text_tmp, *exp_text;
  

  // To manage the dummy case
  if ( (ctext == NULL) || (bytes_left <= 0) || (num_obj <= 0) ) {
    *text = NULL;
    *text_len = 0;
    *decoded = 0;
    return ;
  }

  // To manage a decompression point not aligned to a codeword
  if (((*ctext) & (0x80)) == 0){
    fprintf(stderr,"Error: not a tagged byte (HDecompress_nextBlock_tokens_spaceless)\n");
    exit(-1);  }

  // Determines the window to expand (spaces are not alphanumeric tokens....)
  HDecompress_nextBlock_tokens(ctext,bytes_left,num_obj, 
			      &text_tmp,&text_tmp_len,decoded,console);

  // Reinsert the spaces
  HDecompress_expandSpaces(ctext,*decoded, &exp_text, &exp_len,console);


  // Decompresses just the succeeding exp_len bytes that contain the "num_object" words
  HDecompress_nextBlock_bytes(exp_text,exp_len,text,text_len,console);
 
}





/** Decompresses the piece of compressed text that precedes "ctext"
 * and for a specified number of alphanumeric tokens. The Console data
 * structure must be initialized. It does not work well on the spaceless
 * model  (ie. first reserved byte is 1). 
 * In this case  the decompressed string is without spaces between 
 * alphanumeric tokens. Here, the programmer must use the proper procedure.
 *
 * @param ctext pointer to the block to be decompressed.
 * @param bytes_left number of bytes yet available to be decompressed to the left of "ctext".
 * @param num_obj number of alphanumeric tokens to be decompressed.
 * @param text address of the pointer to the decompressed block (allocated by the procedure).
 * @param text_len address of the integer that will contain the length of the decompressed block.
 * @param decoded address of the integer indicating the number of decompressed bytes.
 * @param console pointer to the data structure that contains all the compression infos.
 */

void HDecompress_previousBlock_tokens(char *ctext, int bytes_left, int num_obj, 
				     char **text, int *text_len, int *decoded,
				     Console *console)
{
  int token_len;
  char *token;
  int taggedcwlen;


  // To manage the dummy case
  if ( (ctext == NULL) || (bytes_left <= 0) || (num_obj <= 0) ) {
    *text = NULL;
    *text_len = 0;
    *decoded = 0;
    return ;
  }

  // To manage a decompression point not aligned to a codeword
  if (((*ctext) & (0x80)) == 0){
    fprintf(stderr,"Error: not a tagged byte (HDecompress_previousBlock_tokens)\n");
    exit(-1);  }

  for(*decoded=0; (*decoded < bytes_left)  && (num_obj > 0); *decoded += taggedcwlen){
    
    // get next token from the compressed text
    HToken_decompressPrevious(ctext - *decoded, bytes_left - *decoded,
			     &token,&token_len,&taggedcwlen,console);
    if (isalnum(token[0])){
      num_obj--;
    }

    free(token);
  }

  if (((*(ctext - *decoded)) & (0x80)) ==0){
    fprintf(stderr,"Error: not a tagged byte (HDecompress_previousBlock_Tokens)\n");
    exit(-1);  }

  // Decompresses just the previous "i" bytes that contain the "num_object" words
  HDecompress_nextBlock_bytes(ctext - *decoded, *decoded,text,text_len,console);
 
}




/** Decompresses the piece of compressed text that precedes "ctext"
 * and for a specified number of alphanumeric tokens in the Spaceless
 * model. The Console data structure 
 * must be initialized. The programmer must ensure that the
 * compression type is 'spaceless' (ie. first reserved byte is 1). 
 *
 * @param ctext pointer to the block to be decompressed.
 * @param bytes_left number of bytes yet available to be decompressed to the left of "ctext".
 * @param num_obj number of alphanumeric tokens to be decompressed.
 * @param text address of the pointer to the decompressed block (allocated by the procedure).
 * @param text_len address of the integer that will contain the length of the decompressed block.
 * @param decoded address of the integer indicating the number of decompressed bytes.
 * @param console pointer to the data structure that contains all the compression infos.
 */

void HDecompress_previousBlock_tokens_spaceless(char *ctext, int bytes_left, int num_obj, 
					       char **text, int *text_len, int *decoded,
					       Console *console)
{
  int text_tmp_len, exp_len;
  char *text_tmp, *exp_text;
  

  // To manage the dummy case
  if ( (ctext == NULL) || (bytes_left <= 0) || (num_obj <= 0) ) {
    *text = NULL;
    *text_len = 0;
    *decoded = 0;
    return ;
  }

  // To manage a decompression point not aligned to a codeword
  if (((*ctext) & (0x80)) == 0){
    fprintf(stderr,"Error: not a tagged byte (HDecompress_previousBlock_tokens_spaceless)\n");
    exit(-1);  }


  // Determines the window to expand (spaces are not alphanumeric tokens....)
  HDecompress_previousBlock_tokens(ctext,bytes_left,num_obj, 
				  &text_tmp,&text_tmp_len,decoded,console);

  // Reinsert the spaces
  HDecompress_expandSpaces(ctext-*decoded, *decoded, &exp_text, &exp_len, console);


  // Decompresses just the succeeding exp_len bytes that contain the "num_object" words
  HDecompress_nextBlock_bytes(exp_text,exp_len,text,text_len,console);
 
}








/* ------------------------------------------------------ */
/* -------------------  Token operations ---------------- */
/* ------------------------------------------------------ */




/** Computes the position in the dictionary of the token having the
 * passed plain codeword. Recall that the dictionary is ordered by
 * decreasing codeword length and alphabetically among equally-long
 * codewords.
 *
 * @param plaincw plain codeword of a token.
 * @param plaincwlen length in bits of that codeword.
 * @param cano pointer to a data structure containing the canonical infos.
 * @param token_rank pointer to the integer that will contain the token position in the dictionary.
 */
void HToken_RankFromPlainCw(int plaincw, int plaincwlen, Canonical *cano, 
				int *token_rank) 
{  
  int len = plaincwlen / 7;

  if((plaincwlen % 7) != 0){
    fprintf(stderr,"Error: HToken_RankFromPlainCw - length no 7-multiple\n");
    exit(-1);  }

  *token_rank = cano->offsetcw[len] + plaincw - cano->firstcw[len];

}


/** Computes the position in the dictionary of the token having the
 * passed tagged codeword. Recall that the dictionary is ordered by
 * decreasing codeword length and alphabetically among equally-long
 * codewords.
 * 
 * @param taggedcw tagged codeword of a token.
 * @param taggedcwlen length in bytes of that codeword.
 * @param cano pointer to a data structure containing the canonical infos.
 * @param token_rank pointer to the integer that will contain the token position in the dictionary.
 */

void HToken_RankFromTaggedCw(int taggedcw,int taggedcwlen,Canonical *cano,int *token_rank) 
{  
  int plaincw,plaincwlen;

  if((taggedcwlen > 4) || (taggedcwlen < 1)){
    fprintf(stderr,"Error: HToken_RankFromTaggedCw - length no in [1,4]\n");
    exit(-1);  }

  HCodeword_PlainFromTagged(taggedcw,taggedcwlen,&plaincw,&plaincwlen);

  if((plaincwlen % 7) != 0){
    fprintf(stderr,"Error: length plain no 7-multiple (HToken_RankFromTaggedCw)\n");
    exit(-1);  }

  HToken_RankFromPlainCw(plaincw,plaincwlen,cano,token_rank);

}



/** Determines the token corresponding to a given tagged codeword.
 *
 * @param taggedcw tagged codeword of a token.
 * @param taggedcwlen length in bytes of that codeword.
 * @param cano pointer to a data structure containing the canonical infos.
 * @param dict pointer to a Dictionary data structure.
 * @param token pointer to the retrieved token.
 * @param token_len pointer to its computed length. 
 */

void HToken_fromTaggedCw(int taggedcw, int taggedcwlen, 
			Canonical *cano, Dictionary *dict,
			char **token, int *token_len) 
{  
  int token_rank;
  int token_startpos;

  HToken_RankFromTaggedCw(taggedcw,taggedcwlen,cano,&token_rank);

  token_startpos = dict->start_pos[token_rank];
  if ((token_startpos < 0) || (token_startpos >= dict->length)){
    fprintf(stderr,"Error: token_startpos computation (HToken_fromTaggedCw)\n");
    exit(-1);  }  
    
  *token_len=1; // jump the first char which does exist

  while(dict->content[token_startpos + (*token_len)] != '\n') 
    (*token_len)++;

  *token = (char *) malloc(*token_len);
  memcpy(*token,dict->content + token_startpos, *token_len);

  if(*token == NULL) {
    fprintf(stderr,"Error: memory allocation for token (HToken_FromTaggedCw)\n");
    exit(-1);  }  

}


/** Decompresses the next token starting from the byte position
 * indicated by s. It also computes the length of this token and the
 * length (in bytes) of the decompressed codeword. The procedure
 * returns 0 if the operation was applied on <=0 bytes, otherwise it
 * returns 1. 
 *
 * @param s pointer to the byte position where the decompression must start.
 * @param num_byte_left number of bytes to the right of s.
 * @param token address of the string pointer that will contain the decompressed token.
 * @param lentoken address of the integer that will contain the length of the token.
 * @param lencw address of the integer that will contain the byte length of the decompressed codeword.
 * @param console pointer to the data structure containing all the compression infos.
 */

int HToken_decompressNext(char *s, int num_byte_left, char **token, int *lentoken, int *lencw, 
			 Console *console)
{

  int taggedcw;

  if(num_byte_left <= 0) return 0;

  // get the next tagged codeword from the compressed text
  HCodeword_TaggedGetNext(s,num_byte_left,&taggedcw,lencw);

  // compute its token
  HToken_fromTaggedCw(taggedcw,*lencw,&(console->canoinfo),&(console->dictionary),token,lentoken);

  return 1;

}


/** Decompresses the token whose codeword precedes the byte position
 * indicated by s. It also computes the length of this token and the
 * length (in bytes) of the decompressed codeword. The procedure
 * returns 0 if the operation was applied on <=0 bytes, otherwise it
 * returns 1.
 *
 * @param s pointer to the byte position where the decompression must start.
 * @param num_byte_left number of bytes to the left of s.
 * @param token address of the string pointer that will contain the decompressed token.
 * @param lentoken address of the integer that will contain the length of the token.
 * @param lencw address of the integer that will contain the byte length of the decompressed codeword.
 * @param console pointer to the data structure containing all the compression infos.
 */

int HToken_decompressPrevious(char *s, int num_byte_left, char **token, int *lentoken, int *lencw,
			     Console *console)
{

  int taggedcw;

  if(num_byte_left <= 0) return 0;

  // get the previous tagged codeword from the compressed text
  HCodeword_TaggedGetPrevious(s,num_byte_left,&taggedcw,lencw);
  
  // compute its token
  HToken_fromTaggedCw(taggedcw,*lencw,&(console->canoinfo),&(console->dictionary),token,lentoken);

  return 1;

}



/** Returns the length of a token defined as either a sequence of
 * letters and numbers, or as a sequnece of separators different of
 * 'newline', or as a single 'newline'. This distinction is introduced because
 * the Dictionary string is formed by putting one token per line,
 * hence if 'newline' occurs inside a token than problems may arise.
 *
 * @param s pointer the beginning of the next token in the parsed text.
 * @param num_char_left number of chars to be scanned in s.
 * @param len pointer to the integer that will contain the length of the parsed token.
 */

void HToken_getLengthNext(char *s, int num_char_left, int *len)
{

  *len = 0;

  if (num_char_left <= 0) {
    fprintf(stderr,"Error: HToken_getLengthNext\n");
    exit(-1); } 

    
  if (!isalnum(*s)) {   // token is a sequence of separators (not '\n') or a single '\n'

    while ( (*len < num_char_left) && (!isalnum(*s)) && ((*s) != '\n')) {
	(*len)++; s++; }

    // Parsed just one '\n'
    if (*len == 0)  *len = 1;

  } else { // token is a sequence of ALNUM chars

    while ( (*len < num_char_left) && isalnum(*s) ){
      (*len)++;  s++; }
  }

  if(*len <= 0){
    fprintf(stderr,"Error: HToken_getLengthNext\n");
    exit(-1);  }
}



/* ------------------------------------------------------ */
/* -----------------  Codeword operations --------------- */
/* ------------------------------------------------------ */


/** Computes the plain codeword corresponding to an input tagged and
 * byte-aligned codeword.
 *
 * @param taggedcw byte-aligned and tagged codeword to be translated.
 * @param taggedcwlen number of bytes constituting the above taggedcw.
 * @param plaincw pointer to an integer that will contain the
 * computed plain codeword.  
 *@param plaincwlen pointer to the integer
 * that will contain the length in bits of the plain codeword.
 */
void HCodeword_PlainFromTagged(int taggedcw, int taggedcwlen, int *plaincw, int *plaincwlen) 
{  
  *plaincw=0;
  *plaincwlen=0;

  if((taggedcwlen > 4) || (taggedcwlen < 1)){
    fprintf(stderr,"Error: HCodeword_PlainFromTagged - length no in [1,4]\n");
    exit(-1);  }
  
  while (taggedcwlen > 0) {
    *plaincw += (taggedcw & 0x7F) << (*plaincwlen);
    taggedcw >>= 8;
    *plaincwlen += 7; // it is expressed in bits
    taggedcwlen --;  // it is expressed in bytes
  }

}


/** Computes the plain codeword of a token from its rank in the
 * dictionary. If some codeword length is absent then its
 * firstcw[]=+infty.
 *
 * @param token_rank position in the dictionary of the input token (starts from 0).
 * @param cano pointer to a data structure containing the canonical infos.
 * @param plaincw pointer to an integer eventually containing the computed plain codeword.  
 * @param plaincwlen pointer to the integer that will contain the length in bits of the plain codeword.
 */

void HCodeword_PlainFromTokenrank(int token_rank,Canonical *cano,int *plaincw,int *plaincwlen) 
{  
  int i;

  // codewords stored by decreasing length in the dictionary
  for(i=1; (i < 5) && (cano->offsetcw[i] > token_rank); i++)
    ;

  *plaincw = cano->firstcw[i] + (token_rank - cano->offsetcw[i]);
  *plaincwlen= 7 * i;

}




/** Computes the tagged and byte-aligned codeword from its plain
 * counterpart. 
 * 
 * @param plaincw input plain codeword to be translated.
 * @param plaincwlen length in bits of the plain codeword.
 * @param taggedcw pointer to the integer that will contain the output tagged codeword.
 * @param taggedcwlen pointer to the integer that will contain the length in bytes of the tagged codeword.
 */

void HCodeword_TaggedFromPlain(int plaincw, int plaincwlen, int *taggedcw, int *taggedcwlen) 
{  
  *taggedcw=0;
  *taggedcwlen=0;

  if((plaincwlen % 7) != 0){
    fprintf(stderr,"Error: HCodeword_TaggedFromPlain, len %d\n",plaincwlen);
    exit(-1);  }

  while (plaincwlen > 0) {
    *taggedcw |= ((plaincw & 0x7F) << (8 * (*taggedcwlen)));
    (*taggedcwlen)++;
    plaincw >>= 7;
    plaincwlen -=7;
  }
  
  *taggedcw |= (0x80 << (8 * (*taggedcwlen - 1)));
}

/** Computes the byte-aligned and tagged codeword of a token given its
 * rank in the dictionary (starts from 0).  Recall that if some
 * codeword length is absent then its firstcw[]=+infty. 
 *
 * @param token_rank position in the dictionary of a token.
 * @param cano pointer to a data structure containing the canonical infos.
 * @param taggedcw pointer to the integer that will contain the computed tagged codeword.
 * @param taggedcwlen pointer to the integer that will contain the length in byte of the tagged codeword.
 */
void HCodeword_TaggedFromTokenrank(int token_rank,Canonical *cano, int *taggedcw, int *taggedcwlen) 
{  
  int plaincw,plaincwlen;

  HCodeword_PlainFromTokenrank(token_rank,cano,&plaincw,&plaincwlen);

  if((plaincwlen % 7) != 0){
    fprintf(stderr,"Error: HCodeword_TaggedFromTokenrank -length plain no 7-multiple\n");
    exit(-1);  }

  HCodeword_TaggedFromPlain(plaincw,plaincwlen,taggedcw,taggedcwlen);

  if((*taggedcwlen > 4) || (*taggedcwlen < 1)){
    fprintf(stderr,"Error: HCodeword_TaggedFromTokenrank - length tagged no in [1,4]\n");
    exit(-1);  }

}



/** Computes the tagged and byte-aligned codeword for the passed token.
 *
 * @param token pointer to the string containing the input token.
 * @param token_len length of the input token.
 * @param ht pointer to the hash table containing the parsed tokens.
 * @param taggedcw pointer to the integer that will contain the output tagged codeword.
 * @param taggedcwlen pointer to the length in bytes of the computed tagged codeword. 
 */
void HCodeword_TaggedFromToken(char *token,int token_len,HHash_table *ht,int *taggedcw, int *taggedcwlen) 
{  
  Hash_node *p;

  if((p=HHashtable_search(token,token_len,ht)) == NULL){
    fprintf(stderr,"Error: token not found (HCodeword_TaggedFromToken)\n");
    exit(-1);  }  

  *taggedcw = p->codeword;
  *taggedcwlen = p->cw_len;

}


/** Reads a sequence of bytes corresponding to the next tagged
 * codeword in s. Its length, in bytes, is returned in lencw; the
 * codeword is assigned to *taggedcw and allocated into its least
 * significant bytes.
 *
 * @param s string containing the tagged codeword to be read.
 * @param num_byte_left number of bytes remaining in s.
 * @param taggedcw pointer to the integer that will contain the tagged codeword (allocated rightword).
 * @param lencw pointer to the integer containing the codeword length (in bytes).
 */

void HCodeword_TaggedGetNext(char *s, int num_byte_left, int *taggedcw, int *lencw)
{
  
  if (((*s) & 0x80) == 0) { // First byte is not tagged
    fprintf(stderr,"Error: no tagged byte (HCodeword_TaggedGetNext)\n");
    exit(-1); } 

  if (num_byte_left <= 0) {
    fprintf(stderr,"Error: no byte left (HCodeword_TaggedGetNext)\n");
    exit(-1); } 

  *taggedcw = 0;
  *taggedcw |= *s;
  (*lencw)=1;
  s++;

  while ( (*lencw < num_byte_left) && ( ((*s) & 0x80) == 0 )) {
    *taggedcw = ((*taggedcw) << 8) | (*s);
    (*lencw)++;
    s++;
  }

}


/** Returns the length of the tagged codeword starting at s.
 *
 * @param s string containing the tagged codeword to be read.
 * @param num_byte_left number of bytes remaining in s.
 */

int HCodeword_TaggedGetNextLength(char *s, int num_byte_left)
{
  
  int lencw;

  if (((*s) & 0x80) == 0) { // First byte is not tagged
    fprintf(stderr,"Error: no tagged byte (HCodeword_TaggedGetNextLength)\n");
    exit(-1); } 

  if (num_byte_left <= 0) { // No bytes left
    fprintf(stderr,"Error: no byte left (HCodeword_TaggedGetNextLength)\n");
    exit(-1); } 

  lencw=1;
  s++;  // skip the tagged byte

  while ( (lencw < num_byte_left) && ( ((*s) & 0x80) == 0 )) {
    lencw++;
    s++;
  }

  return lencw;
}


/**  Reads the sequence of bytes preceding the one pointed to by s and
 * corresponding to the previous tagged codeword. The codeword length
 * is returned in lencw (it is smaller than 4); the codeword is
 * assigned to *taggedcw and allocated into its least significant
 * bytes. 
 *
 * @param s pointer to a char position, we need to read backward from it.
 * @param num_byte_left number of bytes remaining to the left of s.
 * @param taggedcw pointer to the integer that will contain the tagged codeword (allocated rightword).
 * @param lencw pointer to the integer containing the codeword length (in bytes).
 */

void HCodeword_TaggedGetPrevious(char *s, int num_byte_left, int *taggedcw, int *lencw)
{
  
  if (((*s) & 0x80) == 0) { // First byte is not tagged
    fprintf(stderr,"Error: no tagged byte (HCodeword_TaggedGetPrevious)\n");
    exit(-1); } 

  if (num_byte_left <= 0) {
    fprintf(stderr,"Error: no byte left (HCodeword_TaggedGetPrevious)\n");
    exit(-1); } 

  *taggedcw = 0;
  *lencw = 0;
  s--; // go to the previous byte

  while ( (*lencw < num_byte_left) && ( ((*s) & 0x80) == 0 )) {
    *taggedcw = (*taggedcw) | ((*s) << 8 * (*lencw));
    (*lencw)++;
    s--;
  }

  // the tagged byte
  *taggedcw = (*taggedcw) | ((*s) << 8 * (*lencw));
  (*lencw)++;

}




/** Transforms a (tagged) codeword into a string of bytes (ended by \0). 
 *
 * @param taggedcw the input tagged codeword.
 * @param taggedcwlen its length in bytes.
 * @param s memory address of the pointer to the output string.
 */
void HCodeword_tostring(int taggedcw, int taggedcwlen, char **s) 
{  

  int i = taggedcwlen;
  *s = (char *) malloc((taggedcwlen+1) * sizeof(char));


  if((*s == NULL) || (taggedcwlen < 1) || (taggedcwlen > 4)){
    fprintf(stderr,"Error: HCodeword_tostring\n");
    exit(-1);  }

  
  // Copy the codeword rightward
  
  while (i >= 0) {
    (*s)[taggedcwlen-i]=(unsigned char) ((taggedcw >> (8 * (i-1))) & 0xFF);
    i--;
  }  

  // End the string properly
  (*s)[taggedcwlen]='\0';

}



/* ---------------------------------------------------- */
/* -------------------  Hash Table    ----------------- */
/* ---------------------------------------------------- */



/** Initialize the hash table according to the number of estimated
 *   tokens; the load factor is set to 10 and the lists are managed
 *   via the MTF-rule. 
 *
 * @param ht pointer to an (empty) hash table.
 * @param n estimated number of items to be inserted.
 */

void HHashtable_init(HHash_table *ht, int n)
{
  
  int i;

  ht->size  = (int) 1.2 * n + 13;     // Load factor ~ 80%

  ht->table = (Hash_nodeptr_array) malloc(ht->size * sizeof(Hash_node *));

  ht->card = 0;
  if (ht->table == NULL) {
    fprintf(stderr,"Fatal Error: Hash table allocation\n");
    exit(-1); }

  for (i=0; i < ht->size; i++)
    ht->table[i] = NULL;
}


/** Print the content of the hash table ht: token, length, count
 * occurrences, tagged cw and its length.
 *
 * @param ht pointer to the hash table to be printed out.
 */

void HHashtable_print(HHash_table *ht)
{  
  int i,plaincw,lcw;
  Hash_node *p;

  printf("\n\n================== Hash Table ====================\n");
  printf("Table size = %d, number of stored objects = %d\n\n",ht->size,ht->card);

  for(i=0; i<ht->size;i++){
    for(p = ht->table[i]; p ; p = p->next){
      printf("token = \"");
      HPrint_string(p->str,p->len_str);
      printf("\"");
      HCodeword_PlainFromTagged(p->codeword,p->cw_len,&plaincw,&lcw);
      printf(" tokenlen = %d count = %d tagged cw = %x cw_len %d\n", 
	     p->len_str, p->count_occ,plaincw,lcw);
    }
  }
}



/** Computes the hash value for the given string. The function was
    proposed by Ramakrishnais and Zobel in a paper appeared in:
    Int. Conf. on DB Systems for advanced applications, 1997.
 * @param s input string whose hash value has to be computed.
 * @param len length of s.
 * @param ht pointer to an hash table. 
 */
int HHashtable_func(char *s, int len, HHash_table *ht)
{   
  register int hfn;
  int hfi;
  int table_size = ht->size;

  hfn = 11;
  for (hfi=0; hfi<len ; hfi++)
    hfn = hfn ^ ((hfn<<5) + (hfn>>2) + (unsigned char) *s++);
  hfn = abs(hfn % table_size);
  return(hfn);
}


/** Searches for the given string into the passed hash table (NULL if not).
 * @param s string to be searched.
 * @param slen length of s (to manage also the NULL char).
 * @param ht pointer to the hash table to be searched.
 */
Hash_node *HHashtable_search(char *s, int slen, HHash_table *ht)
{   
  Hash_node *hsp;  

  for (hsp = ht->table[HHashtable_func(s,slen,ht)]; hsp; hsp = hsp->next)
    if ((slen == hsp->len_str)  && (memcmp(s,hsp->str,slen) == 0)) 
	return(hsp);
  return((Hash_node *)0);
}



/** Inserts the token in the hash table and returns 1 if new, 0
 *  otherwise; it also updates the counter of occurrences for that
 *  token. 
 *
 * @param s string to be inserted.
 * @param slen length of s (to manage also the NULL char).
 * @param ht pointer to an hash table, it will be updated.
 */
int HHashtable_insert(char *s, int slen, HHash_table *ht)   
{   
  int hiv;
  Hash_node *hip;
  Hash_node *table_head;
  Hash_node *hip_pred;
  char *token;

  hiv = HHashtable_func(s,slen,ht);  // compute the hash value
  hip = HHashtable_search(s,slen,ht);   // check string occurrence

  if (hip) {
    hip->count_occ += 1;
    
    // move-to-front the searched token
    if (ht->table[hiv] != hip){

      // Find hip predecessor which surely does exist
      // because hip is not the head of the list
      for (hip_pred = ht->table[hiv]; hip_pred->next != hip; 
	   hip_pred = hip_pred->next) ;

      hip_pred->next = hip->next;   // jump hip in the old list
      table_head = ht->table[hiv];   // save the old head of the list-entry
      ht->table[hiv] = hip;        // hip is the new head of the list
      hip->next = table_head;         
    }

    return(0);

  } else {    //------ The token is new -------
    hip = (Hash_node *) malloc(sizeof(Hash_node));

    token = malloc(slen);
    memcpy(token,s,slen);

    if (hip == NULL){
      fprintf(stderr,"Error: Insert hash table\n");
	  exit(-1); } 

    hip->len_str = slen;
    hip->count_occ = 1;
    hip->str = token;
    hip->next = ht->table[hiv];
    ht->table[hiv] = hip;
    ht->card += 1;
    return(1);
  }
}


/** Frees all elements of a hashtable. After this call, ht is an empty,
 * uninitialized HHash_table.
 * 
 * @param ht pointer to a HHash_table
 */
void HHashtable_clear(HHash_table *ht)
{
  int i;

  for ( i = 0; i < ht->size; i++ ) {
    Hash_node *hn = ht->table[i];
    while ( hn ) {
      Hash_node *toFree = hn;
      hn = hn->next;
      free(toFree->str);
      free(toFree);
    }
  }
  ht->card = 0;
  ht->size = 0;
  free(ht->table);
}


/** Computes an index data structure for the dictionary tokens, based
 *  on a hash table and the canonical info; each entry of the hash
 *  table contains also the corresponding tagged and byte-aligned
 *  codeword and its length (as #bytes). 
 *
 * @param dict pointer to a Dictinary data structure (initialized).
 * @param cano pointer to the data structure containing the canonical infos (initialized).
 * @param ht pointer to an (empty) hash table, it will be initialized.
 */

void HHashtable_fromdict(Dictionary *dict, Canonical *cano, HHash_table *ht)
{

  int token_len,i,j,l;
  char *token;
  Hash_node *p;


  // Creates the hash table for the tokens and their infos
  HHashtable_init(ht,dict->num_tokens);

  // Initializes the data structures
  for(i=0,j=0; i<dict->length; j++){

    i++;
    while (dict->content[i] != '\n')
      i++;
    // this is the token
    token_len = i -  dict->start_pos[j];
    token = dict->content + dict->start_pos[j];  // no memory allocation

    // Update the hash table
    HHashtable_insert(token,token_len,ht);
    if((p=HHashtable_search(token,token_len,ht)) == NULL){
      fprintf(stderr,"Error: token not found (HHashtable_fromdict)\n");
      exit(-1);  }

    HCodeword_TaggedFromTokenrank(j,cano,&(p->codeword),&l);
    p->cw_len = l; //since we store #bytes

    i++;
  }

  if(j != dict->num_tokens) {
    fprintf(stderr,"Error: num tokens in dictionary (HHashtable_fromdict)\n");
    exit(-1);  }

}


/* ------------------------------------------------------ */
/* ---------------------  Dictionary  ------------------- */
/* ------------------------------------------------------ */

/** Computes from a given string containing the tokens separated by
 *   'newline' the dictionary data structure (the last token is also
 *   ended by 'newline'). The string must be kept since the CONTENT
 *   field refers to it.
 *
 * @param s string containing the tokens separated by 'newline' (even the last one).
 * @param slen length of the string s.
 * @param stokens number of tokens in the string s.
 * @param dict pointer to the Dictionary data structure to be built.
 */
void HDictionary_fromstring(char *s, int slen, int stokens, Dictionary *dict)
{

  int i,j;


  dict->length = slen;
  dict->num_tokens=stokens;

  dict->content = s;
  if(dict->content == NULL) {
    fprintf(stderr,"Error: memory allocation (HDictionary_fromstring)\n");
    exit(-1);  }  

  // Creates the array for the starting positions
  dict->start_pos = (int *) malloc(dict->num_tokens * sizeof(int));
  if(dict->start_pos == NULL) {
    fprintf(stderr,"Error: memory allocation (HDictionary_fromstring)\n");
    exit(-1);  }


  // Initializes the data structures
  for(i=0,j=0; i<dict->length; j++){
    dict->start_pos[j] = i;
    i++;
    while (dict->content[i] != '\n')
      i++;
    i++;
  }

  if(j != dict->num_tokens) {
    fprintf(stderr,"Error: num tokens in dictionary (HDictionary_fromstring)\n");
    exit(-1);  }

}

/** Computes the dictionary data structure from the leaves of a
 *  Huffman tree, they are ordered by codeword length and
 *  alphabetically.
 *
 * @param tree pointer to an array of Hash_node pointers corresponding to the Huffman tree leaves.
 * @param num_tokens number of non-dummy leaves.
 * @param dict pointer to the Dictionary data structure to be built.
 */
void HDictionary_fromtree(Hash_nodeptr_array tree, int num_tokens, Dictionary *dict)
{

  int i,j,k;


  dict->length = 0;
  dict->num_tokens=num_tokens;
  
  for(i=0; i<dict->num_tokens; i++)
    dict->length +=  tree[i]->len_str + 1;

  dict->content = (char *) malloc( (dict->length + 1) * sizeof(char));

  if(dict->content == NULL) {
    fprintf(stderr,"Error: memory allocation (HDictionary_fromtree)\n");
    exit(-1);  }  

  for(i=0,j=0; i<dict->num_tokens; i++){
    for(k=0; k < tree[i]->len_str; k++) 
      dict->content[j++] = tree[i]->str[k]; 
    dict->content[j++] = '\n';
  }
  dict->content[j]='\0';

  if(j != dict->length) {
    fprintf(stderr,"Error: dictionary length (HDictionary_fromtree)\n");
    exit(-1);  }

  // Creates the array for the starting positions
  dict->start_pos = (int *) malloc(dict->num_tokens * sizeof(int));

  if(dict->start_pos == NULL) {
    fprintf(stderr,"Error: memory allocation (HDictionary_fromtree)\n");
    exit(-1);  }

  // Initializes the data structures
  for(i=0,j=0; i<dict->length; j++){
    dict->start_pos[j] = i;
    i++;
    while (dict->content[i] != '\n')
      i++;
    i++;
  }

  if(j != dict->num_tokens) {
    fprintf(stderr,"Error: num tokens in dictionary (HDictionary_fromtree)\n");
    exit(-1);  }

}


/** Print the dictionary content and, also, the codewords for every
 *  token according to the format token:cw cwlen if Verbose = 1.
 *
 * @param dict pointer to the Dictionary data structure to print.
 * @param ht pointer to the hash table containing the parsed tokens and their codewords.
 * @param Verbose set to 1 if all those infos must be printed out. 0 just general Dictionary info. 
 */

void HDictionary_print(Dictionary *dict, HHash_table *ht, int Verbose)
{

  int j,plaincw,lcw;
  Hash_node *p;

  printf("\n\n================== Dictionary ====================\n");
  printf("Number of tokens = %d, string length = %d\n\n",dict->num_tokens,dict->length);

  for(j=0;j<dict->num_tokens - 1;j++){
    printf("\n");
    HPrint_string(dict->content+dict->start_pos[j],dict->start_pos[j+1]-dict->start_pos[j]-1);

    if(Verbose) {
      if((p=HHashtable_search(dict->content+dict->start_pos[j],dict->start_pos[j+1]-dict->start_pos[j]-1,ht)) 
	 == NULL){
	fprintf(stderr,"Error: token not found (HDictionary_print)\n");
	exit(-1);  }
      HCodeword_PlainFromTagged(p->codeword,p->cw_len,&plaincw,&lcw);  
      printf(":plain_cw = %x cwlen %d (bits)",plaincw,lcw);
    }
  }

  printf("\n"); // Printing the last dict item
  HPrint_string(dict->content+dict->start_pos[j],dict->length - dict->start_pos[j]-1);

  if(Verbose) {
    if((p=HHashtable_search(dict->content+dict->start_pos[j],dict->length - dict->start_pos[j]-1,ht)) == NULL){
      fprintf(stderr,"Error: token not found (HDictionary_print)\n");
      exit(-1);  }  
    HCodeword_PlainFromTagged(p->codeword,p->cw_len,&plaincw,&lcw);  
    printf(":plain_cw = %x cwlen %d (bits)\n\n",plaincw,lcw);
  }
} 




/* -------------------------------------------------------- */
/* ----------------  Jumpers data structure --------------- */
/* -------------------------------------------------------- */


/** Computes the jumper data structure according to a number of
 * skipped chars. Since the skipping may end up into a word, then we
 * need to store two integers, the offset in the uncompressed text and
 * the offset into the compressed text.
 *
 * @param text pointer to the text string.
 * @param tlen text length.
 * @param ht pointer to the hash table containing the text tokens.
 * @param jump_value number of chars to be skipped by every jumper pointer.
 * @param jumpers pointer to the Jumper data structure to be built.
 */
void HJumpers_fromds(char *text, int tlen, HHash_table *ht, 
		    int jump_value, Jumpers *jumpers)
{
  int text_pos,ctext_pos,i;
  int token_len;
  int taggedcw;
  int taggedcwlen;
  int current_jump;

  jumpers->number = 0;

  if (jump_value != 0) {
    jumpers->number = tlen / jump_value + 1;

  jumpers->text_offsets = (int *) malloc(jumpers->number * sizeof(int));
  jumpers->ctext_offsets =  (int *) malloc(jumpers->number * sizeof(int));

  if((jumpers->text_offsets == NULL) || (jumpers->ctext_offsets == NULL)){
    fprintf(stderr,"Error: memory allocation (HJumpers_fromds)\n");
    exit(-1);  }

  current_jump = 0;

  for(text_pos=0,ctext_pos=0,i=0; (text_pos<tlen) && (i < jumpers->number); ){
    
    if (text_pos >= current_jump){ // candidate to store a jumper
      if ((i==0) || (jumpers->text_offsets[i-1] != text_pos)) { // it is a novel position
      jumpers->text_offsets[i] = text_pos; 
      jumpers->ctext_offsets[i] = ctext_pos;
      i++;
      }
      current_jump += jump_value;
    }

    // Update the current positions
    HToken_getLengthNext(text+text_pos,tlen-text_pos,&token_len);
    HCodeword_TaggedFromToken(text+text_pos,token_len,ht,&taggedcw,&taggedcwlen);

    text_pos += token_len;
    ctext_pos += taggedcwlen;
  }

  jumpers->number = i; // real number of initialized entries

  }

}



/** Serializes the Jumper data structure by storing in the string s
 * the offsets in the text and the offsets in the compressed text,
 * alternatively; it appends to the front of the string s the number
 * of jumpers it has written. 
 *
 * @param s output string.
 * @param jumpers pointer to the Jumper data structure to be serialized.
 */
void HJumpers_tostring(char *s, Jumpers *jumpers)
{

  int i,j;

  if(s == NULL) {
    fprintf(stderr,"Error: memory allocation (HJumpers_tostring)\n");
    exit(-1);  }

  HInt_tostring(s,jumpers->number);
  
  for(i=0,j=0;i<jumpers->number;i++){
    HInt_tostring(s + 4 + j,jumpers->text_offsets[i]);
    j += 4;
    HInt_tostring(s + 4 + j,jumpers->ctext_offsets[i]);
    j += 4;
  }

}


/** Loads the number of jumpers, and then the offsets in the text and
 * the offsets in the compressed text from the string s,
 * alternatively.
 *
 * @param s string containing the jumpers, it starts with their number.
 * @param jumpers pointer to the Jumper data structure to be built.
 */

void HJumpers_fromstring(char *s, Jumpers *jumpers)
{

  int i;

  jumpers->number = HInt_fromstring(s);
  s += 4;

  if (jumpers->number) {
    jumpers->text_offsets = (int *) malloc(jumpers->number * sizeof(int));
    jumpers->ctext_offsets =  (int *) malloc(jumpers->number * sizeof(int));
    
    if((jumpers->text_offsets == NULL) || (jumpers->ctext_offsets == NULL)){
      fprintf(stderr,"Error: memory allocation (HJumpers_fromstring)\n");
      exit(-1);  }
    
    for(i=0;i<jumpers->number;i++){
      jumpers->text_offsets[i]=HInt_fromstring(s);
      s += 4;
      jumpers->ctext_offsets[i]=HInt_fromstring(s);
      s += 4;
    }
  } else {
    jumpers->text_offsets = NULL;
    jumpers->ctext_offsets =  NULL;
  }    
}


/** Computes the text position (counting from 0) corresponding to a
 * given byte position (counting from 0) in the compressed text. We
 * assume that the byte position corresponds to the beginning of a
 * codeword. 
 *
 * @param ctext string containing the compressed text.
 * @param ctext_len length in bytes of the compressed text.
 * @param bytepos byte position corresponding to the beginning of a codeword.
 * @param textpos pointer to the integer that will contain the text position corresponding to bytepos.
 * @param console pointer to the data structure containing all the compression infos.
 */
void Get_charpos_from_bytepos(char *ctext, int ctext_len, int bytepos, int *textpos, Console *console)
{

  int i;
  int current_ctextpos;
  char *token;
  int ltoken,lencw;

  if((ctext_len <= bytepos) || (bytepos < 0)){
    fprintf(stderr,"Error: uncorrect bytepos (Get_charpos_from_bytepos)\n");
    exit(-1);  }

  if((ctext[bytepos] & 0x80) == 0){
    fprintf(stderr,"Error: no codeword start (Get_charpos_from_bytepos)\n");
    exit(-1);  }
  

  for(i=0; (i < console->jumpers.number) && (console->jumpers.ctext_offsets[i]<= bytepos); i++) ;
  
  current_ctextpos = console->jumpers.ctext_offsets[i-1]; // it is a codeword start, by hypothesis
  *textpos = console->jumpers.text_offsets[i-1];
  
  for( ; current_ctextpos < bytepos; ) {

    HToken_decompressNext(ctext+current_ctextpos,ctext_len-current_ctextpos+1,
			 &token,&ltoken,&lencw,console);
    current_ctextpos += lencw;
    *textpos += ltoken;

    free(token);
  }

  if(current_ctextpos != bytepos){
    fprintf(stderr,"Error: no codeword start (Get_charpos_from_bytepos)\n");
    exit(-1);  }


}


/** Computes the byte position (counting from 0) in the compressed
 * text corresponding to a given char position (counting from 0) into
 * the uncompressed text. We assume that the text position refers to
 * the beginning of a token. 
 *
 * @param ctext pointer to the compressed text.
 * @param ctext_len length in bytes of the compressed text.
 * @param textpos a text position corresponding to the beginning of a token.
 * @param bytepos pointer to an integer that will contain the byte position to be computed.
 * @param console pointer to the data structure containing all the compression infos.
 */

void Get_bytepos_from_charpos(char *ctext, int ctext_len, int textpos, int *bytepos, Console *console)
{

  int i;
  int current_textpos;
  char *token;
  int ltoken,lencw;


  if(textpos < 0){
    fprintf(stderr,"Error: negative text pos (From_textpos_to_ctextpos)\n");
    exit(-1);  }

  
  for(i=0; (i < console->jumpers.number) && (console->jumpers.text_offsets[i]<= textpos); i++) ;
  
  current_textpos = console->jumpers.text_offsets[i-1]; // it is a token start, by hypothesis
  *bytepos = console->jumpers.ctext_offsets[i-1];
  
  for( ; (current_textpos < textpos) && (*bytepos < ctext_len); ) {

    HToken_decompressNext(ctext + *bytepos, ctext_len - *bytepos + 1,&token,&ltoken,&lencw,console);
    current_textpos += ltoken; 
    *bytepos += lencw;

    free(token);
  }

  if(textpos != current_textpos){
    fprintf(stderr,"Error: no token beginning (From_textpos_to_ctextpos)\n");
    exit(-1);  }


}







/* ---------------------------------------------------------- */
/* ----------------  Canonical data structure --------------- */
/* ---------------------------------------------------------- */


/** Computes the canonical huffman infos and updates the hash table
 *   by setting for each token its byte-aligned and tagged codeword
 *   and its codeword length in bytes. 
 *
 * @param tree_array pointer to the array of huffman tree leaves, kept as pointers to Hash nodes.
 * @param cano pointer to the data structure that will contain the canonical information.
 * @param ht pointer to the hash table containing all the tokens (this is updated with taggedcw 
 * and their lengths. 
 */
  
void HCanonical_fromtree(Hash_nodeptr_array tree_array, Canonical *cano, HHash_table *ht)
{
  int i,j,tmp;
  int numcw[5];
  int HT_leaves;
  int HT_dummy_leaves, HT_nondummy_leaves;
  

  HT_dummy_leaves = 128 - (ht->card % 127);
  HT_nondummy_leaves = ht->card;
  HT_leaves = HT_nondummy_leaves + HT_dummy_leaves;

  /* Compute offset, numcw and first codeword per level */
  for(i=4;i>0;i--){
    numcw[i] = 0;
    cano->offsetcw[i] = 0;
    cano->firstcw[i] = 0;
  }
  
  // Recall that the leaves due to the sorting are allocated in
  // the part of the array Huff_root + [0...HT_leaves-1].
  // We count the number of leaves occurring at each level of the
  // tree.  Levels are counted from 1.

  for(i=0;i<HT_nondummy_leaves;i++)
    numcw[tree_array[i]->cw_len]++;

  // maximum length of the canonical code (as #7bit groups)
  for(cano->max_cwlen = 4 ; numcw[cano->max_cwlen] == 0; (cano->max_cwlen)--)
    ;


  // Compute the offsets and the first codewords.  Since we
  // counted only NON_dummy leaves, we assume that the dummys are
  // located leftward in the tree. Hence we set
  // firstcw[max_cwlen] = #dummy so that when summing the value
  // of count_depth[max_cwlen] we have exactly the #leaves at
  // the maximum level.

  cano->offsetcw[cano->max_cwlen] = 0;
  cano->firstcw[cano->max_cwlen] = HT_dummy_leaves;  

  for(i=cano->max_cwlen - 1;i>0;i--){
    cano->offsetcw[i] = cano->offsetcw[i+1]+numcw[i+1];
    cano->firstcw[i] = (cano->firstcw[i+1]+numcw[i+1]) >> 7; // fan-out = 128
  }

  // initializes the number of coded items
  cano->num_tokens=ht->card;


  // Two simple tricks.  
  //
  // We reset the first codeword of max length, thus pushing
  // the dummy leaves rightward in their level. There is no
  // problem in the decompression because we consider always
  // differences between firstcw and ranks.  And we assign
  // offsetcw[] = HT_leaves+1 to set that some cwlen is not
  // occurring

  cano->firstcw[cano->max_cwlen] = 0;  
  for(i=4;i>0;i--)
    if (numcw[i] == 0) cano->offsetcw[i] = HT_leaves+1;

  // Initializes the field codeword into the tree leaves, which are ht->card
  for(i=0,j=cano->max_cwlen; i<cano->num_tokens; i++){

    // Moves the codeword length
    if((j>1) && (cano->offsetcw[j-1]<=i))   j--;

    // Set properly the codeword and its length in bytes
    HCodeword_TaggedFromPlain(cano->firstcw[j]+(i-cano->offsetcw[j]),j*7,&(tree_array[i]->codeword),&tmp);
    tree_array[i]->cw_len = j;
  }

}




/** Computes the canonical infos from a string, assuming the sequence
 * max_cwlen,firstcw[1],offsetcw[1], .., firstcw[4],offsetcw[4]. Since
 * they are just 9 integers (36 bytes) we do not use any succinct
 * encoding nor summarization.  
 *
 * @param s string containing the serialized canonical information.
 * @param cano pointer to the data structure where this information will be loaded.
 */

void HCanonical_fromstring(char *s, Canonical *cano)
{

  int i;

  cano->max_cwlen = HInt_fromstring(s);
  s += 4;

  for(i=1;i<5;i++){
    cano->firstcw[i]=HInt_fromstring(s);
    s += 4;
    cano->offsetcw[i]=HInt_fromstring(s);
    s += 4;
  }

}

/** Stores the canonical infos into a (preallocated) string s,
 *  assuming the sequence max_cwlen,firstcw[1],offsetcw[1], ....,
 *  firstcw[4],offsetcw[4]. Since they are just 9 integers we do not
 *  use any succinct encoding nor summarization. The procedure reads
 *  just the first 36 bytes from s.  
 *
 * @param s preallocated string of length at least 36 chars, here the canonical infos will be stored.
 * @param cano pointer to the data structure containing the infos to be written in s.
 */

void HCanonical_tostring(char *s, Canonical *cano)
{

  int i;

  HInt_tostring(s,cano->max_cwlen);
  s += 4;

  for(i=1;i<5;i++){
    HInt_tostring(s,cano->firstcw[i]);
    s += 4;
    HInt_tostring(s,cano->offsetcw[i]);
    s += 4;
  }

}




/* ------------------------------------------------------------ */
/* ---------------  Routines for Huffman Tree   --------------- */
/* ------------------------------------------------------------ */


/** Uses a hash table containing all the parsed tokens, an array of
 *  pointers to the the leaves of the Huffman tree (to be
 *  constructed), and an integer indicating how many leaves should be
 *  formed including the dummy ones, whose count_occ is set to 0 and
 *  str = NULL.
 *
 * @param tree pointer to an array of Hash_node pointers (it will be initialized). 
 * @param num_leaves total number of tree leaves, including the dummy ones.
 * @param ht pointer to the hash table containing the parsed tokens.
 *
 */
void Hufftree_createLeaves(Hash_nodeptr_array tree, int num_leaves, HHash_table *ht)
{
  Hash_node *p;
  int i,j;

  // Initialize each Huffman leaf with an hash_table entry
  for(i=0,j=0; i<ht->card;j++){
    for(p = ht->table[j]; p ; p = p->next)
      tree[i++] = p;
  }
  // Append the dummy leaves: str = NULL and #occ = 0
  for(i=ht->card; i<num_leaves; i++){
    if ((tree[i] = (Hash_node *) malloc(sizeof(Hash_node))) == NULL){
      fprintf(stderr,"Error: Hufftree_createLeaves\n");
      exit(-1); 
    } else {
      tree[i]->str = NULL;
      tree[i]->len_str = 0;
      tree[i]->count_occ = 0;
    }
  }
     
  // Sort the leaves for decreasing frequency
  qsort(tree,num_leaves,sizeof(Hash_node *),HSort_for_freq);

}


/** Builds an Huffman tree with fan-out 128 starting from the array of
 * leaves (pointers to Hash_nodes) stored in work_area. The algorithm
 * uses the trick if Kunth which exploits to queues kept sorted by
 * increasing frequency of nodes. The leaves are initially allocated
 * to the beginning of work_area; processed nodes are moved to the end
 * of work_area. The procedure eventually returns the position in
 * work_area where the root of the built Huffman tree is allocated. 
 *
 * @param work_area an array of pointers to Hash_nodes, the ones containing the tokens.
 * @param tot_nodes overall number of nodes in the Huffman tree.
 * @param leaves overall number of leaves in the Huffman tree.
 */

int Hufftree_fromLeaves(Hash_nodeptr_array work_area, int *tot_nodes, int leaves)
{
  int num_leaves,i,count,chosen_leaves;
  int processed, head, tail;
  Hash_node *p;
  Hash_nodeptr_array Node_queue; // Queue of internal nodes

  // It always points to the last empty position in the work-area
  // where nodes/leaves will be moved after being processed.
  processed = (leaves + (leaves-1)/127 + 1) -1;
  head = 0;
  tail = 0;
  chosen_leaves = 1;  //chooses from leaves' queue by default
  Node_queue=(Hash_nodeptr_array )malloc(sizeof(Hash_node *)*((leaves-1)/127+1));  

  if(Node_queue == NULL) {
    fprintf(stderr,"Out of mem: Node_queue creation in Hufftree_fromLeaves\n");
    exit(-1); }
  
  for(num_leaves = leaves; (num_leaves > 0) || (head != tail); ){ 
    
    // Select the smallest 128 items as children of the new node
    for(i = 0,count = 0; i < 128; i++) {   
      
      if (head == tail)  // Node_queue is empty
	chosen_leaves = 1; // Extract only leaves

      if(num_leaves == 0) // All leaves have been processed
	chosen_leaves = 0; // Extract from Node_queue

      if( (num_leaves > 0) && (head != tail)) {
	
	// minimize the depth
	if(work_area[num_leaves-1]->count_occ <= Node_queue[head]->count_occ)
	  chosen_leaves = 1;
	else 
	  chosen_leaves = 0;
      }
	
	
      if(chosen_leaves == 0){ // Extract from the Node_queue
	work_area[processed] = Node_queue[head];
	count += Node_queue[head]->count_occ;  
	head++;
	processed--;    
      } else {  // Extract from the leaves
	work_area[processed]=work_area[num_leaves-1];
	count += work_area[num_leaves-1]->count_occ;
	processed--;
	num_leaves--; }
    }
      
    // Allocate the new parent node
    if ((p = (Hash_node *) malloc(sizeof(Hash_node))) == NULL){
      fprintf(stderr,"Error: Token parsing (Hufftree_fromLeaves)\n");
      exit(-1); }
    p->str = NULL;   // it is an internal node
    p->len_str = 0;
    p->count_occ = count;
    p->codeword = 0;
    p->cw_len = 0;
    p->next = (Hash_node *) (processed + 1); //pointer to the leftmost child
    
    // Insertion of the new node in its correct position, if not root	
    if((num_leaves > 0) || (head != tail))
      Node_queue[tail++] = p; 
    else 
      work_area[processed] = p; // it is the root
  }
  
  *tot_nodes = leaves + (leaves-1)/127+1 - processed;
  free(Node_queue); 
  
  // returns the offset of the root node in the array
  return(processed);
}


/** Computes the codeword lengths by visiting top-down the Huffman
 * tree. The depth of a node is propagated to its children,
 * incremented by one; these children are stored contiguously and the
 * position of the leftmost child is kept in the field "next" of the
 * Hash_node pointed to by a tree node. At the end the leaves are
 * stored beginning at position "root", are sorted for decreasing
 * codeword length and among equally-long cw they are sorted by
 * alphabetic order. The procedure returns the maximum codeword length
 * just computed. 
 *
 * @param tree array of pointer to Hash_nodes containing the parsed
 * tokens and representing the tree leaves. 
 * @param root position in the array "tree" where the root node is allocated.
 * @param tree_size overall number of tree nodes.
 */
int Hufftree_computeCwLen(Hash_nodeptr_array tree, int root, int tree_size)
{   
  int offset_child,isleaf;
  int i,j;
  int max_length;

  max_length =0;

  // Visit the tree top-down <---> visit the array rightward
  for(i=0; i<tree_size; i++) {
    
    // Scan all the 128 children of a node if they do exist
    offset_child = (int) (tree[root+i]->next);  // get back the leftmost child
    
    // Node i is a leaf if it stores a string or is dummy
    isleaf = (tree[root+i]->str != NULL) || (tree[root+i]->count_occ == 0);  
    
    // Compute the depth by incrementing of one the parent's depth
    if (isleaf == 0){
      for(j=0; j<128; j++){ 
	if(tree[offset_child+j]->count_occ == 0){ // dummy leaf
	  tree[offset_child+j]->cw_len = -1; // useful for next sorting step
	} else {
	  tree[offset_child+j]->cw_len = (tree[root+i]->cw_len) + 1;
	}
	
	
	// Keep track of the maximum codeword length
	if(tree[offset_child+j]->cw_len > max_length)
	  max_length = tree[offset_child+j]->cw_len;
      }
      
      tree[root+i]->cw_len = -1;  // internal node: useful for next sorting
    }
  }
  
  // Sort nodes for decreasing cw_len + alphabetically for equal
  // length. Internal nodes and dummy leaves go to the end because of
  // the -1 above.
  qsort(tree+root,tree_size,sizeof(Hash_node *),HSort_for_cwlen);
  
  return max_length;
}


/** Constructs the Huffman tree with a fan-out 128, over the set of
 *  tokens contained into the hash table HT. It returns the pointer to
 *  the array of tree leaves ordered by decreasing codeword length and
 *  alphabetically among equally-long codewords.
 * 
 * @param tree_array_ptr pointer to the Huffman tree to be built.
 * @param ht pointer to the hash table containing the parsed tokens
 * (with their frequency).
 */
  
void Hufftree_build(Hash_nodeptr_array *tree_array_ptr, HHash_table *ht)
{
  Hash_node *p;
  int i,depth;
  int Huff_root;
  int HT_size, HT_leaves;
  int HT_dummy_leaves, HT_nondummy_leaves;
  

  HT_dummy_leaves = 128 - (ht->card % 127);
  HT_nondummy_leaves = ht->card;
  HT_leaves = HT_nondummy_leaves + HT_dummy_leaves;

  depth = 5; // dummy starting value


  // Allocate memory for the tree structure and dummy leaves
  (*tree_array_ptr) = (Hash_nodeptr_array )malloc(sizeof(Hash_node *)*
					  (HT_leaves + (HT_leaves-1)/127+5));

  if(*tree_array_ptr == NULL) {
    fprintf(stderr,"Error: tree_array_ptr (Hufftree_build)\n");
    exit(-1);  }
  
  while (depth > 4) {
    
    // Create the leaves for the canonical tree in [0..HT_leaves] and
    // sort according to decreasing frequency (0 for dummy leaves)
    Hufftree_createLeaves(*tree_array_ptr,HT_leaves,ht);

    // Build the Huffman canonical tree; Huff_root is the pos in the
    // *tree_array_ptr where the tree nodes start. HT_size is the overall
    // numbet of tree nodes.
    Huff_root = Hufftree_fromLeaves(*tree_array_ptr,&HT_size,HT_leaves); 

    // Compute codeword lengths: sorted decreasing + alphabetically,
    // the leaves are allocated in the positions: Huff_root +
    // [0...HT_nondummy_leaves] in *tree_array_ptr
    depth = Hufftree_computeCwLen(*tree_array_ptr,Huff_root,HT_size);

    if(depth > 4) {

      for(i=0;i<HT_nondummy_leaves;i++){
	p = (*tree_array_ptr)[i+Huff_root]; 
	p->count_occ = p->count_occ / 1.618 + 1;  // golden ratio scaling
      }
    }
  }

  // Moves the pointer to the position where leaves are allocated
  *tree_array_ptr += Huff_root;

}


/** Print infos about the Canonical Huffword code. Recall that we have
 * offsetcw[] > num_tokens if the corresponding cwlen is not
 * occurring. 
 *
 * @param cano pointer to the Canonical data structure to print.
 * @param tree_array array of pointers to Hash-nodes corresponding to the Huffman tree leaves.
 * @param Verbose set to 1 to print also the tokens and their codewords.
 */
void Hufftree_print(Canonical *cano, Hash_nodeptr_array tree_array, int Verbose)
{
  int i,j;
  int plaincw,lcw;

  printf("\n\n================== Huffman Tree ====================\n");

  printf("Number of distinct tokens %d\n",cano->num_tokens);
  printf("Max codeword length = %d (in groups of 7 bits)\n",cano->max_cwlen);
  for(i=cano->max_cwlen; i>0;i--)
    printf("Offset[%d] = %d, First_cw[%d] = %x\n",i,cano->offsetcw[i],i,cano->firstcw[i]);
  
  j=cano->max_cwlen;

  if(Verbose) {

    printf("\n ----- tokens -----\n\n");
    for(i=0; i<cano->num_tokens; i++){

      // Moves the codeword length
      if((j>1) && (cano->offsetcw[j-1]<=i))   j--;
      
      // Print useful infos
      printf("token = \"");
      HPrint_string(tree_array[i]->str,tree_array[i]->len_str);
      printf("\"");
      
      HCodeword_PlainFromTagged(tree_array[i]->codeword,tree_array[i]->cw_len,&plaincw,&lcw);
      printf(" token_len = %d, cwlen = %d (bits), untagged_cw = %x\n", tree_array[i]->len_str,plaincw,lcw);
    }
  }
    
}








/* ---------------------------------------------------------- */
/* --------------------  Print functions -------------------- */
/* ---------------------------------------------------------- */


/** Prints the first l chars of the string s by using [x] to denote an
 * unprintable ASCII value x. 
 *
 * @param s string to be printed.
 * @param l length of the prefix to print.
 */
void HPrint_string(char *s, int l)
{  

  int i;

  if (l<0){
    fprintf(stderr,"Error: negative string length (HPrint_string)\n");
    exit(-1); }
    
  for(i=0; i<l; i++){
    if(!isprint(s[i])){ printf("[%d]",s[i]); } //singleton
    else printf("%c",s[i]);
  }
}






















/* ---------------------------------------------------- */
/* -----------------  Basic routines    --------------- */
/* ---------------------------------------------------- */


/** Stores the integer "i" into the first 4 bytes of "s".
 * @param s output string.
 * @param i integer to be serialized
 */
void HInt_tostring(char *s, int i)
{
  int j;

  for(j=3; j>=0; j--)
    s[3-j] = (char) ((i >> (8 * j)) & 0xff);
}

/** Reads the first 4 bytes of string s and interprets them as an integer. 
 * @param s input string.
 */
int HInt_fromstring(char *s) 
{
  int j,num;

  num=0;

  for(j=0; j<4; j++)
    num = (num << 8) |  (s[j] & 0xff);

  return(num);
}


/* ---------------------------------------------------- */
/* ---------------  Sorting routines    --------------- */
/* ---------------------------------------------------- */


/* ---------------- HSort_for_decreasing_freq ---------------- */
int HSort_for_freq(const void *va, const void *vb) {
  return(((Hash_nodeptr_array ) vb)[0]->count_occ - ((Hash_nodeptr_array ) va)[0]->count_occ);
}

/* ---------------- HSort_for_decreasing_cwlen ---------------- */
int HSort_for_cwlen(const void *va, const void *vb) {

  // Sort by codeword length
  if ((((Hash_nodeptr_array ) vb)[0]->cw_len - ((Hash_nodeptr_array ) va)[0]->cw_len) != 0)
    return((((Hash_nodeptr_array ) vb)[0]->cw_len - ((Hash_nodeptr_array ) va)[0]->cw_len));

  // Sort alphabetically the alnum strings, they have the same length
  if(((Hash_nodeptr_array ) vb)[0]->str == NULL)
    return(-1);
  if(((Hash_nodeptr_array ) va)[0]->str == NULL)
    return(1);
  return(memcmp(((Hash_nodeptr_array ) va)[0]->str,((Hash_nodeptr_array ) vb)[0]->str,
		((Hash_nodeptr_array ) va)[0]->len_str));
}


