/*! 
  \file huffw.c

  \author Paolo Ferragina, Dipartimento di Informatica, Pisa (Italy).

  \date November 2002.

  \brief The Huffword byte-aligned and tagged compression
  algorithm. Here every codeword is derived from a Huffman tree of
  fanout 128, by appending to each group of 7-bits a tagging bit which
  is 1 for the first group and 0 for the others. The program allows to
  choose two distinct parsing rules for the dictionary formation. For
  details see [deMoura, Navarro, Ziviani, Baeza-Yates, ACM Trans Info
  Syst, 2000].
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <math.h>


// Including the Huffword Library
#include "HuffwordLib.h"

// The (de)compressor program
int main(int argc, char *argv[])
{
  int getopt(int argc, char * const *argv, const char *options);
  extern char *optarg;
  extern int optind, opterr, optopt;
  int Is_compression,Verbose,Parsing_rule,Ave_jumper,c;
  struct stat sbuf;
  FILE *Input_File, *Output_File;
  char Input_File_name[255], Output_File_name[255];
  char *Input_string, *Output_string;
  int Input_size, Output_size, l;


  /* ---------- Default parameters ----------- */
  Is_compression = 1;	// Compression by default
  Verbose = 0;		// Detail on ops
  Parsing_rule = 0;     // Parsing: Lett+Num or single_separator as token
  Ave_jumper = 0;    // No computation of jumpers
	
  opterr = 0;
  while ((c=getopt(argc, argv, "dvp:j:")) != -1)
    {
      switch (c)
	{
	case 'd':
	  Is_compression=0; break;		
	case 'p':
	  Parsing_rule = atoi(optarg); break;
	case 'v': 
	  Verbose++; break;
	case 'j':
	  Ave_jumper = atoi(optarg); 
	  if(Ave_jumper < 0) {
	    fprintf(stderr,"\n\n Error: Average jump value must be >= 0\n");
	    exit(1);  }
	  break;		
	case '?':
	  fprintf(stderr,"Unknown option: %c -main-\n", optopt);
	  exit(1);
	}
    }
	
  if ((Ave_jumper != 0) && (Parsing_rule != 0)) {
    fprintf(stderr,"\n\n Error: -j option can be used only with -p 0\n");
    exit(1);  }

  // ----------- usage message ---------------
  if (optind == argc){
    fprintf(stderr,
        "This is a Huffword, byte-aligned, and tagged compressor.\n\n");
    fprintf(stderr, "Usage: huffw [-dv -j A -p N] infile\n");
    fprintf(stderr, 
        "Output: a file with extension .hwz\n\n");
    fprintf(stderr," -d\tdecompress\n");
    fprintf(stderr," -j\taverage distance of auxiliary pointers for faster text access (only with p = 0)\n");
    fprintf(stderr," -p N\tparsing rule: 0 = token is (Lett+Numb)+ or singletons [default],\n ");
    fprintf(stderr,"\t\t      1 = like 0 but dropping spaces (Spaceless model)\n");
    fprintf(stderr," -v\tdetails of the computation\n\n");
    exit(0);
  }

  /* -------- Create file names -------- */
  strcpy(Input_File_name,argv[optind]);
  if(Input_File_name == NULL) {
    fprintf(stderr,"\n\n Error: input file name empty\n");
    exit(1);  }


  if (Is_compression){  // Compression

    strcpy(Output_File_name,Input_File_name);
    strcat(Output_File_name, ".hwz");

  } else {           // Decompression

    l=strlen(Input_File_name);
    if (strcmp(strdup(Input_File_name + l - 4),".hwz") != 0) {
      fprintf(stderr,"Fatal Error: Input file name does not end for \".hwz\"\n");
      exit(-1); 

    } 

    // drops the last 4 chars ".hwz"
    strncpy(Output_File_name,Input_File_name,l-4); 
    Output_File_name[l-4] = '\0';
  }


  /* ------------ Open files  ------------------- */
  if ((Input_File=fopen(Input_File_name,"r"))==NULL) {
    fprintf(stderr,"Fatal Error: Input file does not exist\n");
    exit(-1); 
  } 

  // If the file exists, it exits
  if ((Output_File=fopen(Output_File_name,"r")) != NULL) {
    fprintf(stderr,"Attention: Output file does exist, please remove or rename it!\n");
    exit(-1); 
  }

  if ((Output_File=fopen(Output_File_name,"w"))==NULL) {
    fprintf(stderr,"Fatal Error: Error in creating the output file\n");
    exit(-1); 
  } 

  /* --- Mmap the input: either compressed or plain text --- */
  stat(Input_File_name,&sbuf); 
  Input_size = (int)sbuf.st_size;
  if (Input_size == 0){
    fprintf(stderr,"Fatal Error: Input file empty\n");
    exit(-1); 
  } 
  Input_string = (char *) mmap(NULL,Input_size,PROT_READ,MAP_SHARED,
			       fileno(Input_File),0);

  /* --- Compressing or Uncompressing --- */
  if (Is_compression){ 

    if (Parsing_rule == 0)
      Huffw_compress(Input_string,Input_size,Ave_jumper,&Output_string,&Output_size,Verbose);
    else
      Huffw_spaceless_compress(Input_string,Input_size,&Output_string,&Output_size,Verbose);

  } else { 

    // First byte is reserved
    Parsing_rule = (int) *Input_string;

    if (Parsing_rule == 0)
      Huffw_decompress(Input_string,Input_size,&Output_string,&Output_size,Verbose);
    else
      Huffw_spaceless_decompress(Input_string,Input_size,&Output_string,&Output_size,Verbose);

  }

  /* -------- Writing out the result -------- */
  fwrite(Output_string,1,Output_size,Output_File);    

  /* -------- Exiting -------- */
  munmap(Input_string,Input_size);
  fclose(Input_File);
  fclose(Output_File);

  return 0;

}




















